/* auth.c - Client/Server authentication
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#ifdef __MINGW32__
 #include <windows.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <gcrypt.h>
#include <ncurses.h>

#include "types.h"
#include "context.h"
#include "crypto.h"
#include "errors.h"
#include "gnuzza.h"

static void
calc_digest( const u8 *data, size_t dlen, u8 *digest )
{
    GCRY_MD_HD md = NULL;
    
    md = gcry_md_open( GCRY_MD_SHA1, 0 );
    if ( md ) {
        gcry_md_write( md, data, dlen );
        gcry_md_final( md );
        memcpy( digest, gcry_md_read( md, GCRY_MD_SHA1 ), 20 );
        gcry_md_close( md );
    }
} /* calc_digest */

static int
auth_send_request( CHAT_ctx chat, int sock_fd )
{
    u8 cookie[16], digest[20], r[20], s[20], buf[4];
    char hostname[128];
    int rc = 0, nlen = 0, len = 0;

    rc = gethostname( hostname, sizeof(hostname)-1 );
    if ( rc == -1 ) {
        sock_error( "gethostname" );
        return set_lasterr( CCERR_GENERAL );
    }
    gcry_randomize( cookie, 16, GCRY_STRONG_RANDOM );
    calc_digest( cookie, 16, digest );
    rc = pk_sign( chat->pk, digest, 20, r, s );
    if ( rc )
        return rc;

    len = strlen( hostname );
    rc = chat_write( chat, PACKET_AUTH, sock_fd, hostname, len, &nlen );
    if ( !rc ) {
        len = strlen( chat->nick );
        rc = chat_write( chat, PACKET_AUTH, sock_fd, chat->nick, len, &nlen );
    }
    if ( !rc )
        rc = chat_write( chat, PACKET_AUTH, sock_fd, cookie, 16, &nlen );
    if ( !rc )
        rc = chat_write( chat, PACKET_AUTH, sock_fd, r, 20, &nlen );
    if ( !rc )
        rc = chat_write( chat, PACKET_AUTH, sock_fd, s, 20, &nlen );
    if ( !rc ) {
        long timestamp = time( NULL );
        chat->seqno = timestamp;
        buf[0] = timestamp >> 24;
        buf[1] = timestamp >> 16;
        buf[2] = timestamp >>  8;
        buf[3] = timestamp      ;
        rc = chat_write( chat, PACKET_AUTH, sock_fd, buf, 4, &nlen );
    }
    
    return rc;
} /* auth_send_request */

static int
auth_check_signature( CHAT_ctx chat, const char *nick,
                      const u8 *digest, const u8 *r,
                      const u8 *s, int *ret_sigstat )
{
    public_key_s *key = NULL;
    int rc = 0;

    if ( opt.disable_pk ) {
        *ret_sigstat = SIG_OK;
        return set_lasterr( CCERR_SUCCESS );
    }

    if ( !key ) {
        char *keyfile = key_file_get( "key.db" );
        if ( !keyfile ) {
            *ret_sigstat = SIG_NOKEY;
            return set_lasterr( CCERR_SUCCESS );
        }
        rc = pubkey_get_byname( keyfile, nick, &key );
        gcry_free( keyfile );
        if ( rc ) {
            *ret_sigstat = SIG_NOKEY;
            return set_lasterr( CCERR_SUCCESS );
        }
        chat_pk_init( chat, key );
    }
    rc = pk_verify( key, r, s, digest, 20 );
    if ( !rc )
        *ret_sigstat = SIG_OK;
    else if ( rc == CCERR_NOKEY )
        *ret_sigstat = SIG_NOKEY;
    else
        *ret_sigstat = SIG_BAD;
    return set_lasterr( CCERR_SUCCESS );
} /* auth_check_signature */
    
static int
auth_recv_request( CHAT_ctx chat, int sock_fd, int *ret_sigstat )
{
    u8 cookie[16], r[20], s[20], digest[20], buf[4];
    char nick[128], host[256];
    int type = 0, nlen = 0;
    int rc = 0;
    
    rc = chat_read( chat, &type, sock_fd, host, sizeof host-1, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        host[nlen] = '\0';
        wprintw( msg, "IDENT_CLIENT %s\n", host );
        wrefresh( msg );
        if ( nlen < sizeof(chat->peer) )
            strcpy( chat->peer, host );
    }
    else
        goto fail;
    
    rc = chat_read( chat, &type, sock_fd, nick, sizeof nick-1, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        nick[nlen] = '\0';
        wprintw( msg, "RECEIVED_NICKNAME %s\n", nick );
        if ( nlen > 0 )
            chat_set_nick( chat, nick, 1 );
    }
    else
        goto fail;

    rc = chat_read( chat, &type, sock_fd, cookie, 16, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        calc_digest( cookie, nlen, digest );
        rc = chat_read( chat, &type, sock_fd, r, 20, &nlen );
        if ( !rc )
            rc = chat_read( chat, &type, sock_fd, s, 20, &nlen );
        if ( !rc )
            rc = auth_check_signature( chat, nick, digest, r, s, ret_sigstat );
    }
    if ( !rc ) {
        rc = chat_read( chat, &type, sock_fd, buf, 4, &nlen );
        if ( !rc )
            chat->seqno = (buf[0] << 24)|(buf[1] << 16)|(buf[2] << 8)|buf[3];
    }
    
fail:
    return rc;
} /* auth_recv_request */

int
auth_send_response( CHAT_ctx chat, int sock_fd, int sigstat )
{
    char hostname[128], sstat[1];
    int len = 0, nlen = 0;
    int rc = 0;

    if ( gethostname( hostname, sizeof hostname-1 ) == -1 ) {
        sock_error( "gethostname" );
        return set_lasterr( CCERR_GENERAL );
    }
    len = strlen( hostname );
    rc = chat_write( chat, PACKET_AUTH, sock_fd, hostname, len, &nlen );
    if ( !rc ) {
        len = strlen( chat->nick );
        rc = chat_write( chat, PACKET_AUTH, sock_fd, chat->nick, len, &nlen );
    }
    if ( !rc ) {
        sstat[0] = sigstat;
        rc = chat_write( chat, PACKET_AUTH, sock_fd, sstat, 1, &nlen );
    }
    
    return rc;
} /* auth_send_response */

int
auth_recv_response( CHAT_ctx chat, int sock_fd )
{
    char host[128], nick[128], sigstat[1];
    int rc = 0, type = 0, nlen = 0;

    rc = chat_read( chat, &type, sock_fd, host, sizeof host-1, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        host[nlen] = '\0';
        wprintw( msg, "IDENT_SERVER %s\n", host );
        wrefresh( msg );
        if ( nlen < sizeof(chat->peer) )
            strcpy( chat->peer, host );
    }
    else goto fail;
    
    rc = chat_read( chat, &type, sock_fd, nick, sizeof nick-1, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        nick[nlen] = '\0';
        wprintw( msg, "RECEIVED_NICKNAME %s\n", nick );
        wrefresh( msg );
        if ( nlen > 0 )
            chat_set_nick( chat, nick, 1 );
    }
    else goto fail;

    rc = chat_read( chat, &type, sock_fd, sigstat, 1, &nlen );
    if ( !rc ) {
        if ( type != PACKET_AUTH )
            return set_lasterr( CCERR_PACKET );
        switch ( sigstat[0] ) {
        case SIG_BAD: return set_lasterr( CCERR_AUTH );
        case SIG_NOKEY: return set_lasterr( CCERR_NOKEY );
        }
    }
    
fail:
    return rc;
} /* auth_recv_response */

int
chat_auth_init( CHAT_ctx chat, int sock_fd )
{
    int rc = 0, sigstat = 0;
    const char *str_sigstat[] = {
        "Good signature.",
        "WARNING: BAD signature.",
        "HINT: public key to verify the signature is missing.",
    };
    
    if ( !chat->is_server ) {
        rc = auth_send_request( chat, sock_fd );
        if ( rc ) {
            wprintw( msg, "auth: `%s'\n", cc_strerror( rc ) );
            wrefresh( msg );
            return set_lasterr( CCERR_AUTH );
        }
    }
    else {
        rc = auth_recv_request( chat, sock_fd, &sigstat );
        if ( rc ) {
            wprintw( msg, "auth: `%s'\n", cc_strerror( rc ) );
            wrefresh( msg );
            return set_lasterr( CCERR_AUTH );
        }
    }

    if ( chat->is_server ) {
        if ( sigstat && sigstat < 3 ) {
            wprintw( msg, "%s\n", str_sigstat[sigstat] );
            wrefresh( msg );
        }
        rc = auth_send_response( chat, sock_fd, sigstat );
        if ( rc ) {
            wprintw( msg, "auth (resp): `%s'\n", cc_strerror( rc ) );
            wrefresh( msg );
            return set_lasterr( CCERR_AUTH );
        }
        if ( sigstat )
            return set_lasterr( CCERR_AUTH );
    }
    else {
        rc = auth_recv_response( chat, sock_fd );
        if ( rc ) {
            wprintw( msg, "auth (resp): `%s'\n", cc_strerror( rc ) );
            wrefresh( msg );
            return set_lasterr( CCERR_AUTH );
        }
    }
    return set_lasterr( CCERR_SUCCESS );
} /* chat_auth_init */







