/* config.c - Config file support.
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <gcrypt.h>
#include <ncurses.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif

#include "types.h"
#include "context.h"
#include "gnuzza.h"
#include "errors.h"

#ifdef __MINGW32__
# undef mkdir
# define mkdir( a, b ) mkdir( a )
#endif

static char*
home_dir_get(  void )
{
    char *home = NULL, *dir = NULL;
    const char *s = "/.gnuzza";

    home = getenv( "HOME" );
    if ( home ) {
        dir = gcry_calloc( 1, strlen( home ) + strlen( s ) + 1 );
        if ( !dir )
            return NULL;
        strcpy( dir, home );
        strcat( dir, s );
    }
    
    return dir;
} /* home_dir_get */

int
home_dir_create( void )
{
    struct stat statbuf;
    char *dir = NULL;

    #ifdef __MINGW32__
    return set_lasterr( CCERR_SUCCESS );
    #endif
    
    dir = home_dir_get( );
    if ( dir && stat( dir, &statbuf ) != -1 ) {
        gcry_free( dir );
        if ( S_ISDIR( statbuf.st_mode ) )
            return set_lasterr( CCERR_SUCCESS );
    }
    if ( mkdir( dir, 0700 ) == -1 ) {
        gcry_free( dir );
        return set_lasterr( CCERR_FILE );
    }
    gcry_free( dir );
    
    return set_lasterr( CCERR_SUCCESS );
} /* home_dir_create */

char*
key_file_get( const char *name )
{
    char *dir = NULL, *file = NULL;

    #ifdef __MINGW32__
    return gcry_strdup( name );
    #endif
    
    dir = home_dir_get( );
    if ( dir == NULL )
        return NULL;
    file = gcry_calloc( 1, strlen( dir ) + strlen( name ) + 2 );
    if ( !file ) {
        gcry_free( dir );
        return NULL;
    }
    strcpy( file, dir );
    strcat( file, DIRSEP_S );
    strcat( file, name );
    gcry_free( dir );

    return file;
} /* key_file_get */

static FILE*
config_file_open( void )
{
    FILE *fp;
    char *home = NULL, *p = NULL;
	
    home = getenv( "HOME" );
    if ( home == NULL ) {
        fp = fopen( ".gnuzza.conf", "rb" );
        if ( fp == NULL ) {
            perror( ".gnuzza.conf" );
            return NULL;
        }
        return fp;
    }
    
    p = gcry_malloc( strlen( home ) + 16 );
    if ( !p )
        return NULL;
    strcpy( p, home );
    strcat( p, DIRSEP_S".gnuzza.conf" );
    
    fp = fopen( p, "rb" );
    if ( fp == NULL ) {
        perror( p );
        return NULL;
    }
    gcry_free( p );
    return fp;
} /* config_file_open */

void
config_file_parse( void )
{
    FILE *fp = NULL;
    char entry[1024];
    int llen = 0;
	
    fp = config_file_open();
    if ( !fp )
        return; /* the config file is not really needed */

    /* we can't use the gcry_ memory routines here because we can't
       differ if they were created by popt or this function. */
    
    while ( !feof( fp ) ) {
        char *p = fgets( entry, 1024, fp );
        if ( !p )
            break;
        if ( *p == '#' )
            continue; /* skip comment lines */
        if ( strchr( entry, '\n' ) )
            entry[strlen( entry ) - line_len] = '\0';
        llen = strlen( entry );
        if ( llen >= 10 && !memcmp( entry, "disable-pk", 10 ) )
            opt.disable_pk = 1;
        else if ( llen >= 12 && !memcmp( entry, "cipher-algo", 11 ) ) {
            opt.cipher_algo = 1;
            argval_algo = strdup( entry + 12 );
        }
        else if ( llen >= 5 && !memcmp( entry, "nick", 4 ) ) {
            opt.nick = 1;
            argval_nick = strdup( entry + 5 );
        }
        else if ( llen >= 11 && !memcmp( entry, "key-length", 10 ) ) {
            opt.key_length = 1;
            argval_bits = strdup( entry + 11 );
        }
        else if ( llen >= 6 && !memcmp( entry, "color", 5 ) ) {
            opt.color = 1;
            argval_color = strdup( entry + 6 );
        }
        else if ( llen >= 11 && !memcmp( entry, "enable-sftp", 11 ) )
            opt.enable_sftp = 1;
        else if ( llen >= 10 && !memcmp( entry, "ppp-speed", 9 ) )
            opt.speed = atol( entry + 10 );
    }
    
    fclose( fp );
} /* config_file_parse */






