/* context.c - Chat context manipulation
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <gcrypt.h>
#include <ncurses.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif

#include "context.h"
#include "types.h"
#include "crypto.h"
#include "gnuzza.h"
#include "errors.h"

int
chat_new( CHAT_ctx *ret_ctx )
{
    CHAT_ctx c;
	
    c = gcry_calloc( 1, sizeof *c );
    if ( !c )
        return set_lasterr( CCERR_MEMORY );
    c->color = WHITE;
    c->r_color = GREEN;
    if ( ret_ctx )
        *ret_ctx = c;
    
    return set_lasterr( CCERR_SUCCESS );
} /* chat_new */ 

void
chat_cipher_close( GCRY_CIPHER_HD *hd )
{
    if ( *hd ) {
        gcry_cipher_close( *hd );
        *hd = NULL;
    }
} /* chat_cipher_close */
    
void
chat_release( CHAT_ctx c )
{
    if ( !c )
        return;
    chat_cipher_close( &c->srv );
    chat_cipher_close( &c->cli );
    if ( opt.enable_sftp ) {
        chat_cipher_close( &c->srv_ftp );
        chat_cipher_close( &c->cli_ftp );
    }
    dh_release( c->dh ); c->dh = NULL;
    pubkey_release( c->pk ); c->pk = NULL;    
    gcry_free( c );
} /* chat_release */

int
chat_init( CHAT_ctx *c, int server, int cipher_type )
{
    int algo, mode, rc = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
    rc = chat_new( c );
    if ( rc )
        return rc;
    chat_set_server_flag( *c, server );
    mode = GCRY_CIPHER_MODE_CFB;
    switch ( cipher_type ) {
    case CC_3DES_CFB: algo = GCRY_CIPHER_3DES; break;
    case CC_BLOWFISH_CFB: algo = GCRY_CIPHER_BLOWFISH; break;
    case CC_TWOFISH_CFB: algo = GCRY_CIPHER_TWOFISH; break;
    case CC_RIJNDAEL_CFB: algo = GCRY_CIPHER_AES; break;
    case CC_CAST5_CFB: algo = GCRY_CIPHER_CAST5; break;
    default: algo = GCRY_CIPHER_3DES; break;
    }
    chat_set_cipher( *c, algo );
    chat_set_cipher_mode( *c, mode );
    
    return 0;
} /* chat_init */

int
chat_pk_init( CHAT_ctx c, public_key_s *pk )
{
    if ( opt.disable_pk )
        return set_lasterr( CCERR_SUCCESS );
    if ( !c || !pk )
        return set_lasterr( CCERR_INV_VALUE );
    c->pk = pk;
    wprintw( msg, "PUBLIC_KEY `%s' pgp-sign-dss %d bits keyID 0x%08X\n",
             c->nick? c->nick : "", gcry_mpi_get_nbits( pk->p ), pk->keyid );
    wrefresh( msg );
    return set_lasterr( CCERR_SUCCESS );
} /* chat_pk_init */

int
chat_dh_init( CHAT_ctx c, int bits )
{
    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
    return dh_init( &c->dh, bits );
} /* chat_dh_init */

int
chat_dh_handshake( CHAT_ctx c, int sock )
{
    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
    return dh_handshake( c->dh, c, sock );
} /* chat_dh_handshake */

void
chat_dh_delete_secrets( CHAT_ctx c )
{
    if ( c && c->dh )
        dh_delete_secrets( c->dh );
} /* chat_dh_delete_secrets */

void
chat_set_server_flag(CHAT_ctx c, int yes)
{
    if ( c )
        c->is_server = yes;
} /* chat_set_server_flag */

void
chat_set_cipher(CHAT_ctx c, int algo)
{
    if ( c )
        c->cipher_algo = algo;
} /* chat_set_cipher */

void
chat_set_cipher_mode(CHAT_ctx c, int mode)
{
    if ( c )
        c->cipher_mode = mode;
} /* chat_set_cipher_mode */

int
chat_set_nick( CHAT_ctx c, const char *nick, int get )
{
    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
    if ( get ) {
        if ( c->r_nick ) {
            gcry_free( c->r_nick );
            c->r_nick = NULL;
        }
        c->r_nick = gcry_strdup( nick );
        if ( c->r_nick == NULL )
            return set_lasterr( CCERR_MEMORY );
    }
    else {
        if ( c->nick ) {
            gcry_free( c->nick );
            c->nick = NULL;
        }
        c->nick = gcry_strdup( nick );
        if ( c->nick == NULL )
            return set_lasterr( CCERR_MEMORY );
    }
    return set_lasterr( CCERR_SUCCESS );
} /* chat_set_nick */

void
chat_set_addr( CHAT_ctx c, const char *addr )
{
    if ( !c )
        return;
    if ( !addr ) {
        memset( c->addr, 0, sizeof c->addr );
        return;
    }
    strcpy( c->addr, addr );
} /* chat_set_device */

void
chat_set_colors( CHAT_ctx c, int local, int peer )
{
    if ( c ) {
        c->color = local;
        c->r_color = peer;
    }
} /* chat_set_colors */






