/* dh-keyex.c - Diffie-Hellman key agreement.
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <gcrypt.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif

#include "prime.h"
#include "errors.h"
#include "context.h"
#include "crypto.h"
#include "types.h"

ciphers_s ciphers[] = {
    { "3DES", CC_3DES_CFB },
    { "BLOWFISH", CC_BLOWFISH_CFB } ,
    { "TWOFISH", CC_TWOFISH_CFB } ,
    { "RIJNDAEL", CC_RIJNDAEL_CFB },
    { "CAST5", CC_CAST5_CFB },
    { NULL, 0 }
};

int
dh_new( DH_STATE *c )
{
    DH_STATE ctx;
	
    ctx = gcry_calloc( 1, sizeof *ctx );
    if ( !ctx )
        return set_lasterr( CCERR_MEMORY );
    if ( c )
        *c = ctx;
    return set_lasterr( CCERR_SUCCESS );
} /* dh_new */

void
dh_mpi_release( GCRY_MPI *ret_mpi )
{
    if ( ret_mpi ) {
        gcry_mpi_release( *ret_mpi );
        *ret_mpi = NULL;
    }
} /* dh_mpi_release */

void
dh_release( DH_STATE c )
{
    if ( c ) {
        dh_mpi_release( &c->g );
        dh_mpi_release( &c->p );
        dh_mpi_release( &c->xy );
        dh_mpi_release( &c->XY );
        dh_mpi_release( &c->skey );
        gcry_free( c );
    }
} /* dh_release */

void
dh_delete_secrets( DH_STATE c )
{
    if ( c ) {
        dh_mpi_release( &c->xy );
        dh_mpi_release( &c->skey );
    }
} /* dh_delete_secrets */

static void
dh_set_bits( DH_STATE c, int bits )
{
    if ( c )
        c->bits = bits? bits: 1024;
} /* dh_set_bits */

int
dh_init( DH_STATE *c, int bits )
{
    int rc = 0;
  
    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
        
    rc = dh_new( c );
    if ( rc )
        return rc;
    dh_set_bits( *c, bits );
    return set_lasterr( CCERR_SUCCESS );
} /* dh_init */

int
dh_set_cipher_prefs( DH_STATE c, ... )
{
    va_list ap, _ap;
    int rc = 1, i, num = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
        	
    va_start( ap, c );
    _ap = ap;
    while ( va_arg(ap, u8) != 0 )
        num++;
    if ( num && ( num <= 5 ) ) {
        for ( i = 0; i < num; i++ )
            c->cipher_prefs[i] = va_arg( _ap, u8 );
        c->ncipher = num;
        rc = 0;
    }
    va_end( ap );
	
    return set_lasterr( rc );
} /* dh_set_cipher_prefs */

int
dh_set_prime_prefs( DH_STATE c, ... )
{
    va_list ap, _ap;
    int rc = 1, i, num = 0;

    if ( !c )
        return CCERR_INV_VALUE;
        	
    va_start( ap, c );
    _ap = ap;
    while ( va_arg(ap, u16) != 0 )
        num++;
    if ( num && ( num <= 5 ) ) {
        for ( i = 0; i < num; i++ )
            c->prime_prefs[i] = va_arg( _ap, u16 );
        c->nprimes = num;
        rc = 0;
    }
    va_end( ap );
	
    return rc;
} /* dh_set_prime_prefs */

int
dh_get_cipher_id( const char *str_algo )
{
    int i = 0;
    int algo = CC_3DES_CFB; /* default algorithm */
	
    for ( i = 0; str_algo && ciphers[i].algo_id; i++ ) {
        if ( !strcmp( ciphers[i].name, str_algo ) )
            algo = ciphers[i].algo_id;
    }
    
    return algo;
} /* dh_get_cipher_id */

const char*
dh_get_cipher_name( int id )
{
    int i = 0;
    const char *s = ciphers[0].name; /* default algorithm */
	
    for ( i = 0; ciphers[i].algo_id; i++) {
        if ( ciphers[i].algo_id == id )
            s = ciphers[i].name;
    }
    return s;
} /* dh_get_cipher_name */

/*
 * The size of g was taken from the GNU TLS project.
 */
static int
dh_set_params( DH_STATE c )
{
    size_t nbytes = 0;
    int rc = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
        	
    if ( c->bits < 1024 || c->bits > 4096 ) {
        printf( "warning, weird prime size detected; corrected.\n" );
        c->bits = 2048;
    }
	
    switch ( c->bits ) {
    case 1024: 
        nbytes = sizeof prime_1024;
        rc = gcry_mpi_scan( &c->p, GCRYMPI_FMT_USG, prime_1024, &nbytes );
        if ( rc )
            return set_lasterr( CCERR_MPI );
        c->g = mpi_set_ui( NULL, 2 );
        break;
		
    case 1536:
        nbytes = sizeof prime_1536;
        rc = gcry_mpi_scan( &c->p, GCRYMPI_FMT_USG, prime_1536, &nbytes );
        if ( rc )
            return set_lasterr( CCERR_MPI );
        c->g = mpi_set_ui( NULL, 5 );
        break;
      
    case 2048:
        nbytes = sizeof prime_2048;
        rc = gcry_mpi_scan( &c->p, GCRYMPI_FMT_USG, prime_2048, &nbytes );
        if ( rc )
            return set_lasterr( CCERR_MPI );
        c->g = mpi_set_ui( NULL, 5 );
        break;
		
    case 3072:
        nbytes = sizeof prime_3072;
        rc = gcry_mpi_scan( &c->p, GCRYMPI_FMT_USG, prime_3072, &nbytes );
        if ( rc )
            return set_lasterr( CCERR_MPI );
        c->g = mpi_set_ui( NULL, 13 );
        break;
		
    case 4096:
        nbytes = sizeof prime_4096;
        rc = gcry_mpi_scan( &c->p, GCRYMPI_FMT_USG, prime_4096, &nbytes );
        if ( rc )
            return set_lasterr( CCERR_MPI );
        c->g = mpi_set_ui( NULL, 5 );
        break;
    }
	
    if ( !c->g || !c->p )
        return set_lasterr( CCERR_MPI );
	
    return set_lasterr( CCERR_SUCCESS );
} /* dh_set_params */

static int
dh_get_x_size( DH_STATE c )
{
    if ( !c )
        return 0;
        
    if ( c->bits < 1024 || c->bits > 4096 ) {
        printf( "warning, weird prime size detected; corrected.\n" );
        c->bits = 2048;
    }

    switch ( c->bits ) {
    case 1024: return 192;
    case 1536: return 208;
    case 2048: return 232;
    case 3072: return 272;
    case 4096: return 312;
    }
    
    return 384; /* should be enough for all primes */
} /* dh_get_x_size */

/*
 * Calculates X = g^x mod prime or Y = g^y mod prime.
 */
static int
dh_calc_xy( DH_STATE c )
{
    int x_size = 0, ec = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
        	
    ec = dh_set_params( c );
    if ( ec )
        return ec;
    x_size = dh_get_x_size( c );
    c->xy = mpi_new( x_size );
    if ( !c->xy )
        return set_lasterr( CCERR_MPI );
    gcry_mpi_randomize( c->xy, x_size, GCRY_STRONG_RANDOM );
    c->XY = mpi_new( c->bits );
    if ( !c->XY )
        return set_lasterr( CCERR_MPI );
		
    mpi_powm( c->XY, c->g, c->xy, c->p );
    dh_mpi_release( &c->g );
    dh_mpi_release( &c->p );
	
    return set_lasterr( CCERR_SUCCESS );
} /* dh_calc_xy */

/*
 * Calculates skey = Y^x mod prime or skey = X^y mod prime.
 */
static int
dh_calc_skey( DH_STATE c )
{	
    int ec = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );
        
    ec = dh_set_params( c );
    if ( ec )
        return ec;
    c->skey = mpi_new( c->bits );
    if ( !c->skey )
        return set_lasterr( CCERR_MPI );
    mpi_powm( c->skey, c->XY, c->xy, c->p );
    dh_mpi_release( &c->p );
    dh_mpi_release( &c->g );
	
    return set_lasterr( CCERR_SUCCESS );
} /* dh_calc_skey */

static int
rnd_seed_create( u32 seqno, const u8 *buf, size_t bufsize, u8 *md )
{
    GCRY_MD_HD hd;
    int dlen = 0;
	
    hd = gcry_md_open( GCRY_MD_RMD160, 0 );
    if ( !hd )
        return set_lasterr( CCERR_GCRYPT );
    dlen = gcry_md_get_algo_dlen( GCRY_MD_RMD160 );
	    
    gcry_md_putc( hd, seqno >> 24 );
    gcry_md_putc( hd, seqno >> 16 );
    gcry_md_putc( hd, seqno >>  8 );
    gcry_md_putc( hd, seqno >>  0 );
    /* also include the size into the hash */
    gcry_md_putc( hd, bufsize >> 24 );
    gcry_md_putc( hd, bufsize >> 16 );
    gcry_md_putc( hd, bufsize >>  8 );
    gcry_md_putc( hd, bufsize >>  0 );
    gcry_md_write( hd, buf, bufsize );
    memcpy( md, gcry_md_read( hd, GCRY_MD_RMD160 ), dlen );
    gcry_md_close( hd );

    return set_lasterr( CCERR_SUCCESS );
} /* rnd_seed_create */

static void
gen_key_material( u8 *md, u8 *rnd )
{
    if ( md && rnd ) {
        int dlen = gcry_md_get_algo_dlen( GCRY_MD_RMD160 );
        gcry_md_hash_buffer( GCRY_MD_RMD160, rnd, md, dlen );
        memcpy( md, rnd, dlen );
    }
} /* gen_key_material */

/*
 * For secure ftp we use two extra ARC4 keys with 160-bits.
 */
static int
init_sftp_keys( CHAT_ctx c, u8 *md, int md_algo )
{
    u8 key_ftpc[20], key_ftps[20];
    int mode = GCRY_CIPHER_MODE_STREAM, dlen = 0, rc = 0;

    gen_key_material( md, key_ftpc );
    gen_key_material( md, key_ftps );
    dlen = gcry_md_get_algo_dlen( md_algo );

    c->srv_ftp = gcry_cipher_open( GCRY_CIPHER_ARCFOUR, mode, 0 );
    if ( !c->srv_ftp )
        return set_lasterr( CCERR_GCRYPT );
    rc = gcry_cipher_setkey( c->srv_ftp, key_ftps, dlen );
    if ( rc ) {
        chat_cipher_close( &c->srv_ftp );
        return set_lasterr( CCERR_GCRYPT );
    }

    c->cli_ftp = gcry_cipher_open( GCRY_CIPHER_ARCFOUR, mode, 0 );
    if ( !c->cli_ftp ) {
        chat_cipher_close( &c->srv_ftp );
        return set_lasterr( CCERR_GCRYPT );
    }
    rc = gcry_cipher_setkey( c->cli_ftp, key_ftpc, dlen );
    if ( rc ) {
        chat_cipher_close( &c->srv_ftp );
        chat_cipher_close( &c->cli_ftp );
        return set_lasterr( CCERR_GCRYPT );
    }
    
    return set_lasterr( CCERR_SUCCESS );
} /* init_sftp_keys */
    
/*
 * Convert the result of the DH key agreement into a server
 * and a client encryption key.
 */
static int
dh_skey_to_cipher_key( CHAT_ctx c, const u8 *buf, size_t bufsize )
{
    u8 md[20], key_srv[40], iv_srv[20], key_cli[40], iv_cli[20];
    int ivlen = 0, keylen = 0;
    int rc = 0;
	
    rnd_seed_create( c->seqno, buf, bufsize, md );
    /* we use 960 bits of entropy */
    gen_key_material( md, key_srv );
    gen_key_material( md, key_srv + 20 );
    gen_key_material( md, iv_srv );
    gen_key_material( md, key_cli );
    gen_key_material( md, key_cli + 20 );
    gen_key_material( md, iv_cli );
    
    if ( opt.enable_sftp ) {
        rc = init_sftp_keys( c, md, GCRY_MD_RMD160 );
        if ( rc )
            return rc;
    }
    
    ivlen = gcry_cipher_get_algo_blklen( c->cipher_algo );
    keylen = gcry_cipher_get_algo_keylen( c->cipher_algo );
    c->srv = gcry_cipher_open( c->cipher_algo, c->cipher_mode, 0 );
    if ( !c->srv )
        return set_lasterr( CCERR_GCRYPT );

    rc = gcry_cipher_setkey( c->srv, key_srv, keylen );
    if ( rc ) {
        chat_cipher_close( &c->srv );
        return set_lasterr( CCERR_GCRYPT );
    }
    rc = gcry_cipher_setiv( c->srv, iv_srv, ivlen );
    if ( rc ) {
        chat_cipher_close( &c->srv );
        return set_lasterr( CCERR_GCRYPT );
    }
	
    c->cli = gcry_cipher_open( c->cipher_algo, c->cipher_mode, 0 );
    if ( !c->cli ) {
        chat_cipher_close( &c->srv );
        return set_lasterr( CCERR_GCRYPT );
    }
    rc = gcry_cipher_setkey( c->cli, key_cli, keylen );
    if ( rc ) {
        chat_cipher_close( &c->srv );
        chat_cipher_close( &c->cli );
        return set_lasterr( CCERR_GCRYPT );
    }
    gcry_cipher_setiv( c->cli, iv_cli, ivlen );
	
    /* burn out all sentensive information */
    memset( key_srv, 0, sizeof key_srv );
    memset( iv_srv, 0, sizeof iv_srv );
    memset( key_cli, 0, sizeof key_cli );
    memset( iv_cli, 0, sizeof iv_cli );
    memset( md, 0, sizeof md );
	
    return set_lasterr( CCERR_SUCCESS );
} /* dh_skey_to_cipher_key */

/*
 * Socket DH key agreement. Each partners wait until both sides
 * have sent their DH values. Then the shared key is calculated.
 */
int
dh_handshake( DH_STATE dh, CHAT_ctx c, int sock_fd )
{
    u8 buf[MAX_MSG_SIZE], t[MAX_MSG_SIZE], bitbuf[2];
    u8 md[20], mdc[20], algo;
    u16 bits = 0;
    int rc = 0, type;
    int tlen, n, len; 
		
    /* exchange key length */
    if ( c->is_server ) {
        rc = chat_write( c, PACKET_PLAIN, sock_fd, dh->prime_prefs, 10, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        rc = chat_read( c, &type, sock_fd, bitbuf, 2, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        bits = (bitbuf[0] << 8) | bitbuf[1];
        if ( bits > 4096 )
            return set_lasterr( CCERR_DHKEX );
        if ( dh->bits != bits )
            dh->bits = bits;
    }
    else {
        int i;
        u16 prefs[5];
		
        rc = chat_read( c, &type, sock_fd, prefs, 10, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        bits = 2048; /* default */
        for ( i = 0; i < 5; i++ ) {
            if ( prefs[i] == dh->bits )
                bits = prefs[i];
        }
        bitbuf[0] = bits >> 8;
        bitbuf[1] = bits;
        rc = chat_write( c, PACKET_PLAIN, sock_fd, bitbuf, 2, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
    }
	
    /* exchange cipher preferences */
    if ( c->is_server ) {
        rc = chat_write( c, PACKET_PLAIN, sock_fd, dh->cipher_prefs, 5, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        rc = chat_read( c, &type, sock_fd, &algo, 1, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        if ( !dh_get_cipher_name( algo ) )
            return set_lasterr( CCERR_DHKEX );
        c->cipher_algo = algo;
    }
    else {
        int i;
        u8 prefs[5];
		
        rc = chat_read( c, &type, sock_fd, prefs, 5, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        algo = CC_3DES_CFB; /* default */
        for ( i = 0;  i < 5; i++ ) {
            if ( prefs[i] == c->cipher_algo )
                algo = prefs[i];
        }
        rc = chat_write( c, PACKET_PLAIN, sock_fd, &algo, 1, NULL );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
    }

    if ( dh_calc_xy( dh ) )
        return set_lasterr( CCERR_DHKEX );
	
    /* exchange shared secret */
    if ( c->is_server ) {
        tlen = sizeof t;
        gcry_mpi_print( GCRYMPI_FMT_USG, t, &tlen, dh->XY );
        rc = chat_write( c, PACKET_MPI, sock_fd, t, tlen, &n ); /* request */
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        rc = chat_read( c, &type, sock_fd, buf, sizeof buf, &n ); /*response*/
        if ( rc || type != PACKET_MPI )
            return set_lasterr( CCERR_DHKEX );
        dh_mpi_release( &dh->XY );
        rc = gcry_mpi_scan( &dh->XY, GCRYMPI_FMT_USG, buf, &n );
        if ( rc )
            return set_lasterr( CCERR_MPI );
    }
    else {
        rc = chat_read( c, &type, sock_fd, buf, sizeof buf, &n ); /*response*/
        if ( rc || type != PACKET_MPI )
            return set_lasterr( CCERR_DHKEX );
		
        tlen = sizeof t;
        gcry_mpi_print( GCRYMPI_FMT_USG, t, &tlen, dh->XY );
        rc = chat_write( c, PACKET_MPI, sock_fd, t, tlen, &n ); /* request */
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        dh_mpi_release( &dh->XY );
        rc = gcry_mpi_scan( &dh->XY, GCRYMPI_FMT_USG, buf, &n );
        if ( rc )
            return set_lasterr( CCERR_MPI );
    }
	
    rc = dh_calc_skey( dh );
    if ( rc )
        return rc;
	
    len = sizeof buf;
    gcry_mpi_print( GCRYMPI_FMT_USG, buf, &len, dh->skey );
    
    /* exchange MD of the session key */
    if ( c->is_server ) {
        /* fixme: even if we use RIPE-MD 160 to generate the key material,
           it might be insecure to exchange the hash in plaintext. we
           should replace this with a signature. */
        gcry_md_hash_buffer( GCRY_MD_SHA1, md, buf, len );
        rc = chat_write( c, PACKET_PLAIN, sock_fd, md, sizeof md, &n );
        if ( rc )
            return set_lasterr( CCERR_DHKEX );
        rc  = chat_read(c, &type, sock_fd, mdc, sizeof mdc, &n );
        if ( rc || type != PACKET_PLAIN || n != 20 )
            return set_lasterr( CCERR_DHKEX );
        if ( memcmp( md, mdc, 20 ) )
            return set_lasterr( CCERR_DHKEX ); /* transmission failure */
    }
    else {
        gcry_md_hash_buffer( GCRY_MD_SHA1, md, buf, len );
        rc = chat_read( c, &type, sock_fd, mdc, sizeof mdc, &n );
        if ( rc || type != PACKET_PLAIN || n != 20 )
            return set_lasterr( CCERR_DHKEX );
        rc = chat_write( c, PACKET_PLAIN, sock_fd, md, sizeof md, &n );
        if ( rc || !n )
            return set_lasterr( CCERR_DHKEX );
        if ( memcmp( md, mdc, 20 ) )
            return set_lasterr( CCERR_DHKEX ); /* transmission failure */
    }
    
    rc = dh_skey_to_cipher_key( c, buf, c->dh->bits / 8 );
    memset( buf, 0, sizeof buf ); /* burn all information */
    if ( rc )
        return rc;
    
    return set_lasterr( CCERR_SUCCESS );
} /* dh_handshake */


