/* gnuzza.c - Gnuzza (a free and secure peer to peer chat app)  
 *   Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#ifndef __MINGW32__
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
#else
 #include <windows.h>
#endif
#include <stdlib.h>
#include <signal.h>
#include <gcrypt.h>
#include <popt.h>
#include <ncurses.h>

#include "context.h"
#include "types.h"
#include "socket.h"
#include "gnuzza.h"
#include "crypto.h"
#include "errors.h"

char *colors[] = {0, "white", "red", "green", "blue", "yellow", "cyan", 0};
option_s opt;

char *argval_addr = "(none)";
char *argval_nick = "(none)";
char *argval_algo = "(none)";
char *argval_bits = "(none)";
char *argval_color = "(none)";
char *argval_debug = "(none)";
char *argval_pk = "(none)";
char *argval_key = "(none)";
char *argval_out = "(none)";
char *argval_fpr = "(none)";
char *argval_speed = "(none)";

static void    
ctrl_handler( int sig )
{
    printf( "\nsignal %d caught.\n", sig );
    tty_deinit_con( );
    exit( 0 );
} /* ctrl_handler */

static void
gnuzza_exit( void )
{
    tty_deinit_con( );
} /* gnuzza_exit */

static void
show_version( void )
{
    const char *s =
        PACKAGE" (CryptChat) "VERSION"\n"
        "Copyright (C) 2002 Timo Schulz\n"
        "This program comes with ABSOLUTELY NO WARRANTY.\n"
        "This is free software, and you are welcome to redistribute it\n"
        "under certain conditions. See the file COPYING for details.\n";
    fprintf( stdout, "%s", s );
} /* show_version */

static void
show_warranty( int quit )
{
    const char *s = 
        "This is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by"
        "\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"	
        "It is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software "
        "Foundation,\n"
        "Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n";
    fprintf( stdout, "%s", s );
    if ( quit )
        exit( 0 );
} /* show_warranty */

static void
setup_cipher_table( DH_STATE dh )
{
    if ( !dh )
        return;
    dh_set_cipher_prefs( dh, CC_3DES_CFB, CC_BLOWFISH_CFB, CC_TWOFISH_CFB,
                         CC_RIJNDAEL_CFB, CC_CAST5_CFB, 0 );
    dh_set_prime_prefs( dh, 1024, 1536, 2048, 3072, 4096, 0 );
} /* setup_cipher_table */

static int
get_color_id( const char *name )
{
    int i = 0;
    const char *s;
    
    for ( i = 0; (s = colors[i]) && name; i++ ) {
        if ( !strcmp( s, name ) )
            return i;
    }
    
    return -1;
} /* get_color_id */

static void
map_colors( const char *param, int *loc_color, int *rem_color )
{
    const char *p;
    char *ptr = NULL;
    size_t length;
    
    if ( !param || !loc_color || !rem_color )
        return;
    
    *loc_color = 0;
    *rem_color = 0;
    p = strchr( argval_color, ':' );
    if ( !p )
        return;
    length = p-argval_color;
    ptr = gcry_malloc( length + 1 );
    if ( !ptr )
        return;
    ptr[length] = '\0';
    strncpy( ptr, argval_color, length );
    *loc_color = get_color_id( argval_color + length + 1 );
    *rem_color = get_color_id( ptr );
    gcry_free( ptr );
} /* map_colors */

void
pubkey_conv( char *keyfile )
{
    char nick[128], pass[128] = {0}, pass2[128] = {0}, yes[8];
    int sec = 0, rc = 0, equal = 0;

    fprintf( stdout, "Does the file contain the secret key? (y/n) " );
    fgets( yes, sizeof yes-1, stdin );
    yes[strlen(yes)-1] = '\0';

    fprintf( stdout, "Please enter your nick: " );
    fgets( nick, sizeof nick-1, stdin );
    nick[strlen(nick)-1] = '\0';

    sec = ( yes[0] == 'y' || yes[0] == 'Y' )? 1 : 0;
    if ( sec ) {
        while ( equal == 0 ) {
            tty_read_passphrase( pass, sizeof pass-1, 1 );
            tty_read_passphrase( pass2, sizeof pass2-1, 2 );
            if ( !strcmp( pass, pass2 ) ) {
                memset( pass2, 0, sizeof pass2 );
                equal = 1;
            }
            else
                printf( "passphrases do not match. try again.\n" );
        }
    }
    rc = pubkey_convert( keyfile, sec, nick, pass );
    if ( rc )
        printf( "convert key `%s'.\n", cc_strerror( rc ) );

    /* burn the passphrase data */
    memset( pass, 0, sizeof pass );
    memset( pass2, 0, sizeof pass2 );
} /* pubkey_conv */
    
int
main( int argc, char **argv )
{
    int c = 0, rc = 0;
    int cipher_algo = CC_3DES_CFB; /* default */
    int key_bits = 2048; /* default */
    int loc_color = -1, rem_color = -1;
    CHAT_ctx chat = NULL;
    public_key_s *pk = NULL;
    const char *id = NULL;
    char *keyfile = NULL;
    poptContext arg;
	
    struct poptOption args[] = {
        /*{"debug", 'd', POPT_ARG_STRING, &argval_debug, 'd',*/
        /*"switch to the debug mode", 0},*/
        {"ppp-speed", 'y', POPT_ARG_STRING, &argval_speed, 'y',
         "how fast the connection is (in kilobytes)", 0},
        {"list-keys", 'l', POPT_ARG_NONE, NULL, 'l',
         "list all available cipher", 0},
        {"cipher-algo", 'a', POPT_ARG_STRING, &argval_algo, 'a',
         "symmetric cipher", 0},
        {"key-length", 'b', POPT_ARG_STRING, &argval_bits, 'b',
         "keylength in bits", 0},
        {"nick", 'n', POPT_ARG_STRING, &argval_nick, 'n',
         "nickname for the session", 0},
        {"color", 'o', POPT_ARG_STRING, &argval_color, 'o',
         "color for the peers (local:remote)", 0},
        {"server", 's', POPT_ARG_STRING, &argval_addr, 's',
         "run in server mode", 0},
        {"client", 'c', POPT_ARG_STRING, &argval_addr, 'c',
         "run client mode", 0},
        {"public-key", 'k', POPT_ARG_STRING, &argval_pk, 'k',
         "public key for authentication", 0},
        {"convert-pk", 't', POPT_ARG_STRING, &argval_key, 't',
         "convert public key to gnuzza format", 0},
        {"disable-pk", 'u', POPT_ARG_NONE, NULL, 'u',
         "disable public key authentication", 0},
        {"enable-sftp", 'g', POPT_ARG_NONE, NULL, 'g',
         "enable secure file transfer", 0},
        {"fingerprint", 'f', POPT_ARG_STRING, &argval_fpr, 'f',
         "show fingerprint of the key", 0},
        {"output", 'p', POPT_ARG_STRING, &argval_out, 'p',
         "use this as the output file", 0},
        {"help", 'h', POPT_ARG_NONE, NULL, 'h', "show help", 0},
        {"warranty", 'w', POPT_ARG_NONE, NULL, 'w',
         "show warranty information", 0},
        {"version", 'v', POPT_ARG_NONE, NULL, 'v', "show version", 0},
        {NULL, 0, 0, NULL, 0, NULL, 0}
    };

    atexit( gnuzza_exit );
    signal( SIGINT, ctrl_handler );
    
    memset( &opt, 0, sizeof opt );
    rc = home_dir_create( );
    if ( rc ) {
        printf( "home directory: %s\n", cc_strerror( rc ) );
        return 0;
    }

    tty_init_con( );
    config_file_parse( );
    
    arg = poptGetContext( PACKAGE, argc, (const char **)argv, args, 0 );
    while ( (c = poptGetNextOpt( arg )) >= 0) {
        switch ( c ) {
        case 'k': opt.pk = 1; break;
        case 'l': opt.list_keys = 1; break;
        case 'c': opt.client = 1; break;
        case 'n': opt.nick = 1; break;
        case 'h': opt.help = 1; break;
        case 's': opt.server = 1; break;
        case 'v': opt.version = 1; break;
        case 'w': opt.warranty = 1; break;
        case 'a': opt.cipher_algo = 1; break;
        case 'b': opt.key_length = 1; break;
        case 'o': opt.color = 1; break;
        case 't': opt.convkey = 1; break;
        case 'u': opt.disable_pk = 1; break;
        case 'g': opt.enable_sftp = 1; break;
        case 'p': opt.output = 1; break;
        case 'f': opt.fpr = 1; break;
        case 'd': opt.debug = 1; log_debug_init( argval_debug ); break;
        case 'y': opt.speed = atol( argval_speed ); break;
        }
    }
    if ( c < -1 ) {
        fprintf( stderr, PACKAGE": `%s': %s\n", 
                 poptBadOption( arg, POPT_BADOPTION_NOALIAS ),
                 poptStrerror( c ) );
        return 1;
    }
    
    if ( opt.server && opt.client ) {
        printf( "can't run in both modes!\n" );
        return 1;
    }
	
    if ( opt.warranty )
        show_warranty( 1 );
    if ( opt.help )
        opt.version = 1;
	
    if ( opt.version ) {
        show_version( );
        if ( opt.help ) {
            fputc( '\n', stdout );
            poptPrintHelp( arg, stdout, 0 );
        }
        else {
            int i = 0;
            fprintf( stdout, "\nAvailable ciphers: " );
            for ( i = 0; ciphers[i].algo_id; i++ )
                printf( "%s ", ciphers[i].name );
            printf( "\n" );
            return 0;   
        }
        return 0;
    }
	
    if ( opt.convkey ) {
        pubkey_conv( argval_key );
        return 0;
    }

    if ( opt.list_keys ) {
        keyfile = key_file_get( "key.db" );
        if ( keyfile ) {
            rc = pubkey_list( keyfile );
            if ( rc )
                printf( "list keys: `%s': %s.\n", keyfile, cc_strerror( rc ) );
            gcry_free( keyfile );
        }
        return 0;
    }

    if ( opt.color ) {
        if ( !strchr( argval_color, ':' ) ) {
            printf( "invalid color mode; must be contain a ':'\n" );
            return 0;
        }
        map_colors( argval_color, &loc_color, &rem_color );
    }
    
    if ( opt.cipher_algo && !opt.server )
        cipher_algo = dh_get_cipher_id( argval_algo );

    if ( opt.key_length && !opt.server ) {
        key_bits = atol( argval_bits );
        if ( key_bits % 128 != 0 ) {
            printf( "%d is not a multiple of 128; round up the size.\n",
                    key_bits );
            if ( key_bits < 2048 )
                key_bits = 1536;
            else
                key_bits = 2048;
        }
    }

    if ( opt.fpr ) {
        public_key_s *dsa = NULL;
        char *hexfpr = NULL;

        keyfile = key_file_get( "key.db" );
        if ( keyfile == NULL ) {
            printf( "could not get keydb name.\n" );
            return 0;
        }
        rc = pubkey_get_byname( keyfile, argval_fpr, &dsa );
        if ( rc ) {
            const char *err = cc_strerror( rc );
            printf( "could not load pubic key `%s': %s.\n", argval_fpr, err );
        }
        hexfpr = pubkey_get_fingerprint( dsa, NULL, 1 );
        if ( hexfpr ) {
            printf( "%s %s\n", dsa->id, hexfpr );
            gcry_free( hexfpr );
        }
        gcry_free( keyfile );
        pubkey_release( dsa );
        return 0;
    }
    if ( strcmp( argval_pk, "(none)") )
        id = argval_pk;
    else if ( opt.nick )
        id = argval_nick;
    if ( !opt.disable_pk && id && opt.client ) {
        keyfile = key_file_get( "key.db" );
        if ( keyfile == NULL ) {
            printf( "could not get keydb file.\n" );
            return 0;
        }
        rc = pubkey_get_byname( keyfile, id, &pk );
        gcry_free( keyfile );
        if ( !rc ) {
            keyfile = key_file_get( "key.sec" );
            rc = seckey_load( pk, keyfile );
            gcry_free( keyfile );
        }
        if ( rc ) {
            const char *err = cc_strerror( rc );
            printf( "could not load public key `%s': %s.\n", id, err );
            pk = NULL;
            return 0;
        }
    }

    initscr( );
    start_color( );
    init_colors( );
    create_main_wnds( );
    wcolor_set( msg, CYAN, NULL );

#ifdef __MINGW32__
    gcry_control( GCRYCTL_DISABLE_SECMEM );
#else
    gcry_control( GCRYCTL_INIT_SECMEM, 16384, 0 );
    gcry_control( GCRYCTL_DISABLE_SECMEM_WARN );
#endif

    if ( opt.server ) {
        wprintw( msg, "Wait for client connection...\n" );
        wrefresh( msg );
    }
    
    if ( !strcmp( argval_addr, "(none)" ) && !opt.server ) {
        wprintw( msg, "No command was choosen!\n" );
        wrefresh( msg );
        sleep( 2 );
        endwin( );
        return 0;
    }
    
    chat_init( &chat, opt.server, cipher_algo ); /* default */
    chat_dh_init( chat, key_bits );
    chat_set_addr( chat, argval_addr );
    if ( opt.server ) {
        chat_set_nick( chat, "server", 0 );
        setup_cipher_table( chat->dh );
    }
    else
        chat_set_nick( chat, "client", 0 );
    if ( opt.nick )
        chat_set_nick( chat, argval_nick, 0 );
    
    if ( opt.color && loc_color != -1 && rem_color != -1 )
        chat_set_colors( chat, loc_color, rem_color );

    if ( opt.client )
        chat_pk_init( chat, pk );
    
    sock_init( );
    main_loop( chat );
    sock_deinit( );
    tty_deinit_con( );
    endwin( );
	
    chat_release( chat );
    if ( opt.debug )
        log_debug_close( );
#ifndef __MINGW32__
    gcry_control( GCRYCTL_TERM_SECMEM );
#endif
    return 0;
} /* main */




