/* keyfile.c - Key database operations
 *  Copyright (C) 2002 Timo Schulz <ts@winpt.org>
 *
 * This file is part of Gnuzza.
 *
 * Gnuzza is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Gnuzza is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <gcrypt.h>
#include <ncurses.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif
#include <time.h>

#include "context.h"
#include "types.h"
#include "errors.h"
#include "gnuzza.h"
#include "crypto.h"

#ifdef __MINGW32__
 char *strsep(char **stringp, const char *delim);
 #define LE 0x0D
#else /* POSIX */
 #define LE 0x0A
#endif

#define BAD -1
#define base64val(c) index64[(unsigned int)(c)]

static char b64chars[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '+', '/'
};

static int index64[128] = {
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,62, -1,-1,-1,63,
    52,53,54,55, 56,57,58,59, 60,61,-1,-1, -1,-1,-1,-1,
    -1, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,-1, -1,-1,-1,-1,
    -1,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,-1, -1,-1,-1,-1
};

/* encode a raw binary buffer to a null-terminated base64 strings */
static void
base64_encode( u8 *out, const u8 *in, size_t len, size_t olen )
{
    if ( !out || !in )
        return;
    
    while ( len >= 3 && olen > 10 ) {
        *out++ = b64chars[in[0] >> 2];
        *out++ = b64chars[((in[0] << 4) & 0x30) | (in[1] >> 4)];
        *out++ = b64chars[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
        *out++ = b64chars[in[2] & 0x3f];
        olen -= 4; len -= 3; in += 3;       
    }
    /* clean up remainder */
    if ( len > 0 && olen > 4 ) {
        u8 fragment = 0;
        *out++ = b64chars[in[0] >> 2];
        fragment = (in[0] << 4) & 0x30;
        if ( len > 1 )
            fragment |= in[1] >> 4;
        *out++ = b64chars[fragment];
        *out++ = (len < 2) ? '=' : b64chars[(in[1] << 2) & 0x3c];
        *out++ = '=';               
    }
    *out = '\0';       
} /* base64_encode */

/* Convert '\0'-terminated base64 string to raw u8 buffer.
      Returns length of returned buffer, or -1 on error. */
static int
base64_decode( char *out, const char *in )
{
    int len = 0;
    u8 digit1, digit2, digit3, digit4;

    if ( !out || !in )
        return -1;

    do {
        digit1 = in[0];
        if ( digit1 > 127 || base64val (digit1) == BAD )
            return -1;
        digit2 = in[1];
        if ( digit2 > 127 || base64val (digit2) == BAD )
            return -1;
        digit3 = in[2];
        if ( digit3 > 127 || ((digit3 != '=')
                              && (base64val (digit3) == BAD) ) )
            return -1;
        digit4 = in[3];
        if ( digit4 > 127 || ((digit4 != '=')
                              && (base64val (digit4) == BAD) ) )
            return -1;
        in += 4;
        /* digits are already sanity-checked */
        *out++ = (base64val(digit1) << 2) | (base64val(digit2) >> 4);
        len++;
        if ( digit3 != '=' ) {
            *out++ = ((base64val(digit2) << 4) & 0xf0)
                | (base64val(digit3) >> 2);
            len++;
            if ( digit4 != '=' ) {
                *out++ = ((base64val(digit3) << 6) & 0xc0) | base64val(digit4);
                len++;                                                   
            }
        }
    } while ( *in && digit4 != '=' );
    return len;       
}/* base64_decode */

static u16
chksum( const u8 *buf, size_t blen )
{
    int i = 0;
    u16 sum = 0;
    
    for ( i = 0; i < blen; i++ ) sum += buf[i];
    return sum;  
} /* chksum */

static void
skip_bytes( FILE *fp, size_t len )
{
    while ( fp && len-- )
        fgetc( fp );
} /* skip_bytes */

static u8*
read_bignum_pgp( const u8 *p, size_t *ret_size )
{
    u16 len = 0;
    u8 *bignum = NULL;
    
    len = (p[0] << 8) | p[1];
    len = (len + 7) / 8;
    if ( len > 2048 )
        return NULL; /* mpi overflow */
    if ( ret_size )
        *ret_size = len + 2;
    bignum = gcry_calloc( 1, len + 3 );
    if ( bignum )
        memcpy( bignum, p, len + 2 );
    return bignum;
} /* read_bignum_pgp */

static int
seckey_cipher( u8 *x, size_t xlen, char *pass, const u8 *salt )
{
    GCRY_MD_HD md;
    GCRY_CIPHER_HD chd;
    size_t dlen = 0;
    u8 key[20];

    md = gcry_md_open( GCRY_MD_SHA1, 0 );
    if ( md == NULL )
        return set_lasterr( CCERR_GCRYPT );
    dlen = gcry_md_get_algo_dlen( GCRY_MD_SHA1 );
    gcry_md_write( md, pass, strlen( pass ) );
    gcry_md_write( md, salt, 8 );
    gcry_md_final( md );
    memcpy( key, gcry_md_read( md, 0 ), dlen );
    gcry_md_close( md );

    chd = gcry_cipher_open( GCRY_CIPHER_ARCFOUR, GCRY_CIPHER_MODE_STREAM, 0 );
    if ( chd == NULL )
        return set_lasterr( CCERR_GCRYPT );
    if ( gcry_cipher_setkey( chd, key, dlen ) ) {
        gcry_cipher_close( chd );
        return set_lasterr( CCERR_GCRYPT );
    }
    gcry_cipher_encrypt( chd, x, xlen, NULL, 0 );
    gcry_cipher_close( chd );

    return set_lasterr( CCERR_SUCCESS );
} /* seckey_cipher */

int
seckey_load( public_key_s *pk, const char *file )
{
    FILE *fp;
    char *inbuf = NULL, *p = NULL;
    u8 *bignum = NULL, outbuf[4096], salt[8];
    int len = 0, encrypted = 0;
    size_t nbytes = 0;
    u16 csum = 0;

    fp = fopen( file, "rb" );
    if ( fp == NULL )
        return set_lasterr( CCERR_GENERAL );
    inbuf = gcry_calloc( 1, 4096 );
    if ( !inbuf )
        return set_lasterr( CCERR_MEMORY );
    p = fgets( inbuf, 4095, fp );
    fclose( fp );
    if ( !p ) {
        gcry_free( inbuf );
        return set_lasterr( CCERR_FILE );
    }
    p = strsep( &inbuf, " " );
    if ( !p )
        return set_lasterr( CCERR_GENERAL );
    if ( strncmp( p, "pgp-sign-dss", 12 ) )
        return set_lasterr( CCERR_PACKET );
    if ( strstr( p, "-enc" ) )
        encrypted = 1;
    
    p = strsep( &inbuf, " " );
    if ( !p )
        return set_lasterr( CCERR_GENERAL );

    len = base64_decode( outbuf, p );
    if ( len == -1 )
        return set_lasterr( CCERR_BASE64 );
    if ( encrypted ) {
        memcpy( salt, outbuf, 8 );
        csum = (outbuf[8] << 8) | outbuf[9];
    }

    bignum = read_bignum_pgp( outbuf + 10, &nbytes );
    if ( bignum == NULL )
        return set_lasterr( CCERR_MPI );
    if ( encrypted ) {
        char pass[128];
        
        tty_read_passphrase( pass, sizeof pass - 1, 0 );
        seckey_cipher( bignum + 2, nbytes - 2, pass, salt );
        if ( csum != chksum( bignum + 2, nbytes - 2 ) ) {
            gcry_free( bignum );
            return set_lasterr( CCERR_PASS );
        }
    }
    if ( gcry_mpi_scan( &pk->x, GCRYMPI_FMT_PGP, bignum, &nbytes ) ) {
        gcry_free( bignum );
        return set_lasterr( CCERR_GCRYPT );
    }
    gcry_free( bignum );
    pk->secret = 1;
    
    return set_lasterr( CCERR_SUCCESS );
} /* seckey_load */

static public_key_s*
pubkey_load( const char *p )
{
    public_key_s *key = NULL;
    u8 *mem = NULL, *bignum = NULL;
    const u8 *ptr = NULL;
    size_t nbytes = 0;
    int len = 0;

    mem = gcry_calloc( 1, 4 * strlen( p ) + 1 );
    if ( !mem )
        return NULL;
    len = base64_decode( mem, p );
    if ( len == -1 ) {
        gcry_free( mem );
        set_lasterr( CCERR_BASE64 );
        return NULL;
    }
    ptr = mem;
    key = gcry_calloc( 1, sizeof *key );
    if ( !key ) {
        gcry_free( mem );
        set_lasterr( CCERR_MEMORY );
        return NULL;
    }
    key->blob = gcry_calloc( 1, len + 1 );
    if ( !key->blob ) {
        gcry_free( mem );
        gcry_free( key->blob ); key->blob = NULL;
        set_lasterr( CCERR_MEMORY );
        return NULL;
    }
    memcpy( key->blob, mem, len );
    key->blen = len;

    key->timestamp = (ptr[4] << 24)|(ptr[5] << 16)|(ptr[6] << 8)|ptr[9];
    ptr += 9; /* skip 9 header bytes */
    if ( (bignum = read_bignum_pgp( ptr, &nbytes )) == NULL ) /* p */
        goto fail;
    if ( gcry_mpi_scan( &key->p, GCRYMPI_FMT_PGP, bignum, &nbytes ) ) {
        set_lasterr( CCERR_GCRYPT );
        goto fail;
    }
    ptr += nbytes;
    gcry_free( bignum ); bignum = NULL;
    
    if ( (bignum = read_bignum_pgp( ptr, &nbytes )) == NULL ) /* q */
        goto fail;
    if ( gcry_mpi_scan( &key->q, GCRYMPI_FMT_PGP, bignum, &nbytes ) ) {
        set_lasterr( CCERR_GCRYPT );
        goto fail;
    }
    ptr += nbytes;
    gcry_free( bignum ); bignum = NULL;
    
    if ( (bignum = read_bignum_pgp( ptr, &nbytes )) == NULL ) /* g */
        goto fail;
    if ( gcry_mpi_scan( &key->g, GCRYMPI_FMT_PGP, bignum, &nbytes ) ) {
        set_lasterr( CCERR_GCRYPT );
        goto fail;
    }
    ptr += nbytes;
    gcry_free( bignum ); bignum = NULL;

    if ( (bignum = read_bignum_pgp( ptr, &nbytes )) == NULL ) /* y */
        goto fail;
    if ( gcry_mpi_scan( &key->y, GCRYMPI_FMT_PGP, bignum, &nbytes ) ) {
        set_lasterr( CCERR_GCRYPT );
        goto fail;
    }
    
    ptr += nbytes; gcry_free( bignum );
    bignum = NULL;
    set_lasterr( CCERR_SUCCESS );
    
fail:
    gcry_free( bignum );
    gcry_free( mem );
    return key;
} /* pubkey_load */

char*
pubkey_get_fingerprint( public_key_s *key, u8 *fprbuf, int set )
{
    GCRY_MD_HD md = NULL;
    char *fprhex = NULL;
    int i = 0, dlen = 0;
    u8 fpr[20];

    if ( !key ) {
        set_lasterr( CCERR_INV_VALUE );
        return NULL;
    }
    md = gcry_md_open( GCRY_MD_SHA1, 0 );
    if ( !md ) {
        set_lasterr( CCERR_GCRYPT );
        return NULL;
    }
    dlen = gcry_md_get_algo_dlen( GCRY_MD_SHA1 );
    gcry_md_write( md, key->blob, key->blen );
    gcry_md_final( md );
    memcpy( fpr, gcry_md_read( md, GCRY_MD_SHA1 ), dlen );
    gcry_md_close( md );
    key->keyid = (fpr[16] << 24) | (fpr[17] << 16) | (fpr[18] << 8) | fpr[19];
    if ( !set ) return 0;
    if ( fprbuf ) {
        memcpy( fprbuf, fpr, 16 );
        return NULL;
    }
    else {
        fprhex = gcry_calloc( 1, 128 );
        if ( fprhex == NULL ) {
            set_lasterr( CCERR_MEMORY );
            return NULL;
        }
        for ( i = 0; i < 20; i++ ) {
            int c = (i != 19)? ':' : ' ';
            sprintf( fprhex + 3*i, "%02X%c", fpr[i], c );
        }
        return fprhex;
    }
    return NULL;
} /* pubkey_get_fingerprint */

void
pubkey_release( public_key_s *key )
{
    if ( key ) {
        dh_mpi_release( &key->p );
        dh_mpi_release( &key->q );
        dh_mpi_release( &key->g );
        dh_mpi_release( &key->y );
        gcry_free( key->blob );
        gcry_free( key->id );
        gcry_free( key );
    }
} /* pubkey_release  */

static int
load_file( const char *file, char **ret_buf )
{
    FILE *fp;
    char *buf = NULL;
    struct stat statbuf;

    fp = fopen( file, "rb" );
    if ( fp == NULL )
        return set_lasterr( CCERR_FILE );
    if ( fstat( fileno( fp ), &statbuf ) == -1 )
        return set_lasterr( CCERR_FILE );
    if ( statbuf.st_size == 0 ) {
        *ret_buf = NULL;
        return set_lasterr( CCERR_INV_VALUE );
    }
    buf = gcry_calloc( 1, statbuf.st_size + 1 );
    if ( !buf )
        return set_lasterr( CCERR_MEMORY );
    fread( buf, 1, statbuf.st_size, fp );
    buf[statbuf.st_size] = '\0';

    if ( fclose( fp ) )
        return set_lasterr( CCERR_FILE );
    if ( ret_buf )
        *ret_buf = buf;
    return set_lasterr( CCERR_SUCCESS );
} /* load_file */

static const char*
strdate( long timestamp )
{
    static char date[32];
    struct tm *tm;

    tm = localtime( &timestamp );
    sprintf( date, "%04d-%02d-%02d",
             tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday );
    return date;
} /* strdate */

static void
set_user_id( public_key_s *pk, const char *id )
{
    char *ptr = strchr( id, LE );
    
    if ( ptr && pk ) {
        pk->id = gcry_calloc( 1, (ptr-id) + 1 );
        if ( pk->id )
            strncpy( pk->id, id, (ptr-id) );
    }
} /* set_user_id */

int
pubkey_list( const char *file )
{
    public_key_s *pk = NULL;
    char *keybuf = NULL, *buf = NULL, *p = NULL;
    int rc = 0, idx = 0;

    rc = load_file( file, &keybuf );
    if ( rc )
        return rc;

    printf( "%s\n", file );
    for ( idx = 0; idx < strlen( file ); idx++ )
        fputc( '-', stdout );
    fputc( '\n', stdout );
    idx = 0;
    
    while ( (p = strsep( &keybuf, " " ) ) ) {
        switch ( idx ) {
        case 0:
            if ( strncmp( p, "pgp-sign-dss", 12 ) ) {
                printf( "invalid key database format.\n" );
                return set_lasterr( CCERR_GENERAL );
            }
            break;
            
        case 1:
            buf = gcry_strdup( p );
            break;
            
        case 2:
            pk = pubkey_load( buf );
            gcry_free( buf ); buf = NULL;
            if ( pk ) {
                set_user_id( pk, p );
                pubkey_get_fingerprint( pk, NULL, 0 );
                printf( "pub %dD/%08lX %s %s\n",
                        gcry_mpi_get_nbits( pk->p ),
                        pk->keyid,
                        strdate( pk->timestamp ),
                        pk->id );
                pubkey_release( pk ); pk = NULL;
            }
            idx = 0;
            break;
        }
        idx++;
    }

    return set_lasterr( CCERR_SUCCESS );
} /* pubkey_list */
    
int
pubkey_get_byname( const char *file, const char *nick, public_key_s **ret_key )
{
    char *keybuf = NULL, *p, *buf = NULL;
    int idx = 0, rc = 0;

    if ( ret_key )
        *ret_key = NULL;
    rc = load_file( file, &keybuf );
    if ( rc )
        return rc;
    
    while ( (p = strsep( &keybuf, " " ) ) ) {
        switch ( idx ) {
        case 0:
            if ( strncmp( p, "pgp-sign-dss", 12 ) ) {
                printf( "invalid key database format.\n" );
                return set_lasterr( CCERR_GENERAL );
            }
            break;
            
        case 1:
            buf = gcry_strdup( p );
            break;

        case 2:
            if ( nick && strstr( p, nick ) ) {
                *ret_key = pubkey_load( buf );
                set_user_id( *ret_key, p );
                gcry_free( buf );
                goto done;
            }
            idx = 0; /* reset */
            gcry_free( buf ); buf = NULL;
            break;
        }
        idx++;
    }

done:
    if ( *ret_key && (*ret_key)->p ) {
        public_key_s *k = *ret_key;
        pubkey_get_fingerprint( k, NULL, 0 );
    }
    else
        return set_lasterr( CCERR_NOKEY );

    return set_lasterr( CCERR_SUCCESS );
} /* pubkey_get_byname */

static int
do_pubkey_conv( FILE *in, FILE *out, int sec, const char *nick, char *pass )
{
    int ctb = 0, type = 0, len = 0, c = 0;
    char *inbuf = NULL, *outbuf = NULL;
    u8 timebuf[4];

    if ( sec && pass && *pass )
        fputs( "pgp-sign-dss-enc ", out );
    else
        fputs( "pgp-sign-dss ", out );
    ctb = fgetc( in );
    if ( !( ctb & 0x80 ) && ctb == '-' ) {
        printf( "invalid OpenPGP packet. Please dearmor the key.\n" );
        return set_lasterr( CCERR_PACKET );
    }
    if ( !( ctb & 0x80 ) ) {
        printf( "invalid OpenPGP packet.\n" );
        return set_lasterr( CCERR_PACKET );
    }
    if ( !(ctb & 0x40) ) {
        type = ctb & 0x3f;
        type >>= 2;
    }
    if ( (ctb & 0x03) == 0 )
        len = fgetc( in );
    else if ( (ctb & 0x03) == 1 ) {
        len = fgetc( in ) << 8;
        len |= fgetc( in );
    }
    else if ( (ctb & 0x03) == 2 ) {
        len = fgetc( in ) << 24;
        len |= fgetc( in ) << 16;
        len |= fgetc( in ) <<  8;
        len |= fgetc( in );
    }
    if ( type == 6 || type == 5 ) { /* public/secret key */
        if ( (c = fgetc( in )) != 4 ) {
            printf( "\nonly version 4 public keys are supported.\n" );
            return set_lasterr( CCERR_PACKET );
        }
        timebuf[0] = fgetc( in );
        timebuf[1] = fgetc( in );
        timebuf[2] = fgetc( in );
        timebuf[3] = fgetc( in );
        if ( (c = fgetc( in )) != GCRY_PK_DSA ) {
            printf( "\nonly DSA v4 public keys are supported.\n" );
            return set_lasterr( CCERR_PACKET );
        }    
    }
    if ( type == 6 ) { /* public key */
        inbuf = gcry_calloc( 1, len + 10 );
        if ( !inbuf )
            return set_lasterr( CCERR_MEMORY );
        outbuf = gcry_calloc( 1, 5 * len + 1 );
        if ( !outbuf ) {
            gcry_free( inbuf );
            return set_lasterr( CCERR_MEMORY );
        }
        inbuf[0] = 0x99; /* fixed header (old: public key) */
        inbuf[1] = len >> 8;
        inbuf[2] = len;
        inbuf[3] = 0x04; /* version */
        inbuf[4] = timebuf[0];
        inbuf[5] = timebuf[1];
        inbuf[6] = timebuf[2];
        inbuf[7] = timebuf[3];
        inbuf[8] = GCRY_PK_DSA; /* pubkey algo */
        len -= 6;
        fread( inbuf + 9, 1, len, in );
        base64_encode( outbuf, inbuf, len + 9, 5 * len );
        fwrite( outbuf, 1, strlen( outbuf ), out );
        gcry_free( inbuf );
        gcry_free( outbuf );
    }
    else if ( type == 5 ) { /* secret key */
        u8 *seckey = NULL, *armkey = NULL, salt[8];
        int i = 0, nbytes = 0, csum = 0;

        /* skip p, q, g, y */
        for ( i = 0; i < 4; i++ ) {
            c = fgetc( in ) << 8;
            c |= fgetc( in );
            nbytes = (c + 7) / 8;
            len = nbytes;
            skip_bytes( in, nbytes );
        }
        if ( (c = fgetc( in )) != GCRY_CIPHER_NONE ) {
            printf( "\nonly unencrypted secret keys are supported.\n" );
            return set_lasterr( CCERR_PACKET );
        }
        c = fgetc( in ) << 8;
        c |= fgetc( in );
        nbytes = (c + 7) / 8;
        seckey = gcry_calloc( 1, nbytes + 13 );
        if ( !seckey )
            return set_lasterr( CCERR_MEMORY );
        armkey = gcry_calloc( 1, 5 * (nbytes+4) + 1 );
        if ( !armkey ) {
            gcry_free( seckey );
            return set_lasterr( CCERR_MEMORY );
        }
        /* The first 12 octets are reserved for a header */
        fread( seckey + 12, 1, nbytes, in );
        csum = chksum( seckey + 12, nbytes );
        gcry_randomize( salt, 8, GCRY_STRONG_RANDOM );
        memcpy( seckey, salt, 8 );
        seckey[8] = csum >> 8;
        seckey[9] = csum;
        seckey[10] = c >> 8;
        seckey[11] = c;
        if ( pass && *pass )
            seckey_cipher( seckey + 12, nbytes, pass, salt );
        base64_encode( armkey, seckey, nbytes + 12, 5 * (nbytes+4) );
        fwrite( armkey, 1, strlen( armkey ), out );
        gcry_free( seckey );
        gcry_free( armkey );
    }
    
    fputc( ' ', out );
    fputs( nick, out );
    fputc( '\n', out );
    
    return set_lasterr( CCERR_SUCCESS );
} /* do_pubkey_conv */

int
pubkey_convert( const char *file, int sec, const char *nick, char *pass )
{
    FILE *in, *out;
    int rc = 0;

    if ( opt.output ) {
        out = fopen( argval_out, "a+b" );
        if ( out == NULL )
            return set_lasterr( CCERR_GENERAL );
    }
    else
        out = stdout;

    in = fopen( file, "rb" );
    if ( in == NULL )
        return set_lasterr( CCERR_GENERAL );
    rc = do_pubkey_conv( in, out, sec, nick, pass );
    fclose( in );
    if ( opt.output )
        fclose( out );
    
    return rc;
} /* pubkey_convert */




