/* main.c - Main chat loop
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <gcrypt.h>
#include <ncurses.h>
#include <sys/stat.h>
#ifdef __MINGW32__
 #include <windows.h>
#else /* POSIX */
 #include <netinet/in.h>
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <pthread.h>
#endif

#include "context.h"
#include "socket.h"
#include "types.h"
#include "errors.h"
#include "gnuzza.h"
#include "crypto.h"
#include "sftp.h"

static int stop = 0;
static char *cli_ip = NULL;

#ifdef __MINGW32__
int
wthread_create( HANDLE *ret_hd, LPTHREAD_START_ROUTINE fnc, void *ctx )
{
    SECURITY_ATTRIBUTES sec_attr;
    HANDLE thread_hd = NULL;
    DWORD tid = 0;
    
    memset( &sec_attr, 0, sizeof sec_attr );
    sec_attr.bInheritHandle = FALSE;
    sec_attr.lpSecurityDescriptor = NULL;
    sec_attr.nLength = sizeof(sec_attr);

    thread_hd = CreateThread( NULL, 0, fnc, ctx, 0, &tid );
    if ( thread_hd == NULL ) {
        wprintw( msg, "Could not spawn thread.\n" );
        wrefresh( msg );
        return set_lasterr( CCERR_INV_VALUE );
    }
    if ( ret_hd )
        *ret_hd = thread_hd;
    
    return set_lasterr( CCERR_SUCCESS );
} /* wthread_create */
#endif

static int
proc_sftp_request( THREAD ctx, const char *buf )
{
    int slen = strlen( buf ) - 1;
    
    if ( !opt.enable_sftp ) {
        wprintw( msg, "SFTP subsystem is disabled.\n" );
        wrefresh( msg );
        return -1;
    }
    buf++;
    if ( slen >= 4 && !strncmp( buf, "send", 4 ) ) {
        if ( slen - 4 <= 0 ) {
            wprintw( msg, "Syntax: send [file].\n" );
            wrefresh( msg );
            return -1;
        }
        else
            ctx->sftp.cmd = SFTP_SEND;
    }
    else {
        wprintw( msg, "Unknown SFTP keyword: %s\n", buf );
        wrefresh( msg );
        return -1;
    }
    
    if ( !ctx->sftp.init ) {
        ctx->sftp.conn.hfd = connect_to_server( ctx->chat->addr, SFTP_PORT );
        if ( ctx->sftp.conn.hfd == -1 ) {
            wprintw( msg, "Peer machine hasn't enabled SFTP service.\n" );
            wrefresh( msg );
            return -1;
        }
        ctx->sftp.init = 1;
    }
    buf += 5;
    gcry_free( ctx->sftp.file );
    ctx->sftp.file = gcry_strdup( buf );
    if ( ctx->sftp.file == NULL )
        return -1;

    return 0;
} /* proc_sftp_request */

#ifdef __MINGW32__
static DWORD WINAPI
#else
static void*
#endif
sftp_client_loop( void *c )
{
    struct sftp_s *ctx = (struct sftp_s *)c;
    struct stat statbuf;
    char *p, *remote = NULL;
    int rc = 0, fsize = 0;

    if ( !ctx->init || !ctx->conn.hfd || !ctx->file )
        goto quit;

    if ( ctx->cmd == SFTP_SEND ) {
        if ( stat( ctx->file, &statbuf ) == -1 ) {
            wprintw( msg, "Could not open file `%s'.\n", ctx->file );
            wrefresh( msg );
            goto quit;
        }
        fsize = statbuf.st_size;
        if ( opt.speed && (fsize / 1024) > (opt.speed * 10) ) {
            wprintw( msg, "WARNING: You try to send a large file. "
                     "This can take a while.\n" );
            wrefresh( msg );
        }
        p = strrchr( ctx->file, DIRSEP_C );
        if ( p )
            remote = p + 1;
        else
            remote = ctx->file;
        rc = sftp_send_file( ctx, ctx->file, remote );
        if ( rc ) {
            const char *s = sftp_strerror( rc );
            wprintw( msg, "Send file: %s: %s.\n", ctx->file, s );
            wrefresh( msg );
        }
        else {
            wprintw( msg, "Send file: %s -> %s: done.\n", ctx->file, remote );
            wrefresh( msg );
        }
    }
    
quit:
#ifdef __MINGW32__
    ExitThread( 0 );
    return FALSE;
#else /* POSIX */
    return NULL;
#endif
} /* sftp_client_loop */

static void
server_callback( const char *file )
{
    wprintw( msg, "Received file: %s.\n", file );
    wrefresh( msg );
} /* server_callback */

#ifdef __MINGW32__
static DWORD WINAPI
#else /* POSIX */
static void*
#endif
sftp_server_loop( void *c )
{
    struct sftp_s *ctx = (struct sftp_s *)c;
    char *cli_ip2 = NULL;
    int fd = -1, rc = 0;

    fd = wait_for_client( NULL, SFTP_PORT, &cli_ip2 );
    if ( fd < 0 ) {
        set_lasterr( CCERR_SOCK );
        return NULL;
    }

    if ( cli_ip && cli_ip2 && !strcmp( cli_ip, cli_ip2 ) ) {
        wprintw( msg, "SFTP connection established.\n" );
        wrefresh( msg );
        gcry_free( cli_ip2 ); cli_ip2 = NULL;
    }
    else {
        wprintw( msg, "ALERT: Somebody conntected to your SFTP port!\n" );
        wprintw( msg, "Connection is closed.\n" );
        wrefresh( msg );
        goto quit;
    }

    #if 0
    sftp_set_debug_file( "srv.dbg" );
    sftp_set_debug_level( 2 );
    #endif
    ctx->conn.hfd = fd;
    
    rc = sftp_proc_packets( ctx, server_callback );
    if ( rc ) {
        wprintw( msg, "SFTP: %s.\n", sftp_strerror( rc ) );
        wrefresh( msg );
    }

quit:
    sock_close( fd );
#ifdef __MINGW32__
    ExitThread( 0 );
    return FALSE;
#else
    return NULL;
#endif
} /* sftp_server_loop */

void
start_sftp_thread( struct sftp_s *ctx )
{
#ifdef __MINGW32__
    wthread_create( &ctx->hd, sftp_client_loop, (void *)ctx );
#else /* POSIX */
    pthread_create( &ctx->hd, NULL, sftp_client_loop, (struct sftp_s *)ctx );
#endif
} /* start_sftp_thread */

#ifdef __MINGW32__
static DWORD WINAPI
#else /* POSIX */
static void*
#endif
input_loop( void *c )
{
    THREAD ctx = (THREAD)c;
    CHAT_ctx chat = ctx->chat;
    char buf[MAX_MSG_SIZE];
    int nwritten, rc = 0;
	
    echo();	
    wprintw( inp, "\nMSG> " );
    wrefresh( inp );
	
    for ( ; ; ) {
        wrefresh( msg );
        wgetnstr( inp, buf, MAX_MSG_SIZE );
    begin:
        if ( !strlen( buf ) ) {
            wprintw( inp, "\nMSG> " );
            wrefresh( inp );
            continue;
        }
        if ( !strncmp( buf, "EXIT", 4 ) || !strncmp( buf, "exit", 4 ) ) {
            strcpy( buf, "\n\n" );
            strcat( buf, chat->nick );
            strcat( buf, " says goodbye and quits the session.\n" );
            if ( chat->is_server )
                strcat( buf, "** Server closed connection **\n" );
            else
                strcat( buf, "** Client closed connection **\n" );
            chat_write_msg( chat, ctx->sock, buf, strlen( buf ), NULL );
            stop = 1;
            break;
        }
        
        if ( *buf == '#' ) {
            rc = proc_sftp_request( ctx, buf );
            if ( !rc )
                start_sftp_thread( &ctx->sftp );
            buf[0] = '\0'; sleep( 1 );
            goto begin;
        }
        
        strcat( buf, "\n" );
        rc = chat_write_msg( chat, ctx->sock, buf, strlen( buf ),  &nwritten );
        if ( rc ) {
            wprintw( msg, "\nChat socket: could not write data.\n" );
            wrefresh( msg );
            break;
        }
        wcolor_set( msg, chat->color, NULL );
        wprintw( msg, "%s: ", chat->nick );
        wprintw( msg, "%s", buf );
        wrefresh( msg );
		
        wprintw( inp, "\nMSG> " );
        wrefresh( inp );
    }
#ifdef __MINGW32__
    ExitThread( 0 );
    return FALSE;
#else /* POSIX */
    return NULL;
#endif
} /* input_loop */

int       
main_loop( CHAT_ctx ctx )
{
    struct timeval tv;
    struct sftp_s *ftp = NULL;
    THREAD th = NULL;
    fd_set rdfs;
    char buf[MAX_MSG_SIZE];
    int rc, nread = 0, count = 0;
    int sock = -1;

    if ( !ctx )
        return set_lasterr( CCERR_INV_VALUE );
        	
    if ( ctx->is_server )
        sock = wait_for_client( ctx->addr, PORT, &cli_ip );
    else
        sock = connect_to_server( ctx->addr, PORT );
    if ( sock == -1 )
        return set_lasterr( CCERR_SOCK );

    th = gcry_calloc( 1, sizeof *th );
    if ( !th )
        return set_lasterr( CCERR_MEMORY );
    
    rc = chat_dh_handshake( ctx, sock );
    if ( rc ) {
        wprintw( msg, "\n%s\n", cc_strerror( rc ) );
        wrefresh( msg );
        goto leave;
    }
    if ( opt.enable_sftp ) {
        ftp = gcry_calloc( 1, sizeof *ftp );
        th->sftp.conn.cli = ftp->conn.cli = ctx->cli_ftp;
        th->sftp.conn.srv = ftp->conn.srv = ctx->srv_ftp;
        th->sftp.conn.is_server = ftp->conn.is_server = ctx->is_server;
    }

    if ( opt.enable_sftp && ctx->is_server )
#ifdef __MINGW32__
        wthread_create( &th->sftp.serv_hd, sftp_server_loop, (void *)ftp );
#else /* POSIX */
    pthread_create( &th->sftp.serv_hd, NULL, sftp_server_loop, ftp );
#endif
    
    wprintw( msg, "CIPHER_ALGO %s/DH %d bits\n", 
             dh_get_cipher_name( ctx->cipher_algo ), ctx->dh->bits );
    
    rc = chat_auth_init( ctx, sock );
    if ( rc ) {
        wprintw( msg, "\n%s\n", cc_strerror( rc ) );
        wrefresh( msg );
        goto leave;
    }
	
    chat_dh_delete_secrets( ctx );
	
    wprintw( msg, "\nGo ahead and type your message. To quit the program "
             "use \"EXIT\" or \"exit\".\n" );
    wrefresh( msg );
	
    th->chat = ctx;
    th->sock = sock;
#ifdef __MINGW32__
    rc = wthread_create( &th->thread_hd, input_loop, (void *)th );
    if ( th->thread_hd == NULL ) {
        gcry_free( th );
        return rc;
    }
#else /* POSIX */
    pthread_create( &th->thread_hd, NULL, input_loop, (THREAD)th );
#endif
	
    while ( !stop ) {
        wcolor_set( msg, CYAN, NULL );
        tv.tv_sec = 2;
        tv.tv_usec = 0;
        FD_ZERO( &rdfs );
        FD_SET( sock, &rdfs );

#ifdef __MINGW32__
        do {
            count = select( sock + 1, &rdfs, NULL, NULL, &tv );
        } while ( count < 0 && WSAGetLastError() == WSAEINTR );
#else /* POSIX */
        do {
            count = select( sock + 1, &rdfs, NULL, NULL, &tv );
        } while ( count < 0 && errno == EINTR );
#endif
        if ( stop )
            break;
        if ( FD_ISSET( sock, &rdfs ) ) {
            rc = chat_read_msg( ctx, sock, buf, MAX_MSG_SIZE, &nread );
            if ( rc || strstr( buf, "closed connection" ) )
                stop = 1;
            else if ( nread == -1 ) {
                const char *s = cc_strerror( rc );
                wprintw( msg, "\nChat socket: could not read: %s.\n", s );
                wrefresh( msg );
                break;
            }
            buf[nread] = '\0';
            wcolor_set( msg, ctx->r_color, NULL );
            wprintw( msg, "%s: ", ctx->r_nick );
            wprintw( msg, "%s", buf );
            wrefresh( msg );
        }
    }

leave:
    shutdown( sock, SHUT_RDWR );
    sock_close( sock );
#ifdef __MINGW32__
    CloseHandle( th->thread_hd );
    th->thread_hd = NULL;
    if ( opt.enable_sftp ) {
        CloseHandle( th->sftp.serv_hd );
        th->sftp.serv_hd = NULL;
    }
#endif
    gcry_free( th );
    
    return rc;
} /* main_loop */
