/* packet.c - Packet handling.
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *
 * This file is part of Gnuzza.
 *
 * Gnuzza is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Gnuzza is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <gcrypt.h>
#include <ncurses.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif

#include "types.h"
#include "context.h"
#include "socket.h"
#include "crypto.h"
#include "errors.h"
#include "gnuzza.h"

static void
encode_packet_size( size_t size, u8 *pkt_size )
{
    if ( pkt_size ) {        
        pkt_size[0] = size >> 24;
        pkt_size[1] = size >> 16;
        pkt_size[2] = size >>  8;
        pkt_size[3] = size >>  0;
    }
} /* encode_packet_len */
                
static void
decode_packet_size( u8 *pkt_size, size_t *ret_size )
{
    const u8 *p = NULL;
  
    if ( !pkt_size )
        return;
    p = pkt_size;
    if ( ret_size )
        *ret_size = (p[0] << 24) | (p[1] << 16) | (p[2] <<  8) | p[3];
} /* decode_packet_len */

static GCRY_CIPHER_HD
get_cipher_hd( CHAT_ctx c, int encrypt )
{
    if ( !c )
        return NULL;

    if ( encrypt )
        return c->is_server ? c->srv : c->cli;
    return !c->is_server ? c->srv : c->cli;
} /* get_cipher_hd */

int
read_packet( CHAT_ctx c, packet_s *pkt, int fd, size_t *r_nread )
{
    u8 pktlen[4];
    int nread;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );

    if ( r_nread )
        *r_nread = 0;
    if ( pkt->data_size > MAX_MSG_SIZE )
        return set_lasterr( CCERR_GENERAL );

    if ( sock_read(fd, &pkt->type, 1) == -1 )
        return set_lasterr( CCERR_SOCK );
    if ( sock_read(fd, pktlen, 4) == -1 )
        return set_lasterr( CCERR_SOCK );
  
    if ( pkt->type != PACKET_PLAIN && pkt->type != PACKET_MPI ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 0 );
        gcry_cipher_decrypt( hd, pktlen, 4, NULL, 0 );
    }
    decode_packet_size( pktlen, &pkt->data_size );
    if ( pkt->data_size > MAX_MSG_SIZE )
        return set_lasterr( CCERR_PACKET );
    if ( pkt->type == PACKET_DATA ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 0 );
        if ( sock_read( fd, pkt->mdc, 20 ) == -1 )
            return set_lasterr( CCERR_SOCK );
        gcry_cipher_decrypt( hd, pkt->mdc, 20, NULL, 0 );
        pkt->use_mdc = 1;
    }
	
    if ( pkt->data_size == 0 )
        return set_lasterr( CCERR_SUCCESS );
    nread = sock_read( fd, pkt->data, pkt->data_size );
    if ( pkt->type != PACKET_PLAIN && pkt->type != PACKET_MPI ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 0 );
        gcry_cipher_decrypt( hd, pkt->data, pkt->data_size, NULL, 0 );
    }
    if ( r_nread )
        *r_nread = nread;
  
    return set_lasterr( CCERR_SUCCESS );
} /* read_packet */

int
write_packet( CHAT_ctx c, packet_s *pkt, int fd, size_t *r_nwritten )
{
    u8 pktlen[4];
    int nwritten = 0;

    if ( !c )
        return set_lasterr( CCERR_INV_VALUE );

    if ( r_nwritten )
        *r_nwritten = 0;
    if ( pkt->data_size == 0 )
        return set_lasterr( CCERR_SUCCESS ); /* nothing to do */
    else if ( pkt->data_size > MAX_MSG_SIZE )
        return set_lasterr( CCERR_GENERAL );
    if ( sock_write( fd, &pkt->type, 1 ) == -1 )
        return set_lasterr( CCERR_SOCK );
    encode_packet_size( pkt->data_size, pktlen );
    if ( pkt->type != PACKET_PLAIN && pkt->type != PACKET_MPI ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 1 );
        gcry_cipher_encrypt( hd, pktlen, 4, NULL, 0 );
    }
  
    if ( sock_write( fd, pktlen, 4 ) == -1 )
        return set_lasterr( CCERR_SOCK );
    if ( pkt->use_mdc ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 1 );
        gcry_cipher_encrypt( hd, pkt->mdc, 20, NULL, 0 );
        if ( sock_write( fd, pkt->mdc, 20 ) == -1 )
            return set_lasterr( CCERR_SOCK );
    }
    
    if ( pkt->type != PACKET_PLAIN && pkt->type != PACKET_MPI ) {
        GCRY_CIPHER_HD hd = get_cipher_hd( c, 1 );
        gcry_cipher_encrypt( hd , pkt->data, pkt->data_size, NULL, 0 );
    }
  
    nwritten = sock_write( fd, pkt->data, pkt->data_size );
    if ( r_nwritten )
        *r_nwritten = nwritten;
	
    return set_lasterr( CCERR_SUCCESS );
} /* write_packet */

int
chat_write_msg( CHAT_ctx c, int fd, const void *buf, size_t buflen,
                size_t *r_nwritten )
{
    packet_s pkt;
    size_t nwritten = 0;
    int rc = 0;
	
    if ( !c || !buf || !buflen )
        return set_lasterr( CCERR_INV_VALUE );

    if ( r_nwritten )
        *r_nwritten = 0;
    memset( &pkt, 0, sizeof pkt );
    pkt.type = PACKET_DATA;
    pkt.data_size = buflen;
    memcpy( pkt.data, buf, buflen );
	
    gcry_md_hash_buffer( GCRY_MD_RMD160, pkt.mdc, buf, buflen );
    pkt.use_mdc = 1;

    rc = write_packet( c, &pkt, fd, &nwritten );
    if ( rc )
        return rc;
    if ( r_nwritten )
        *r_nwritten = nwritten;
    
    return set_lasterr( CCERR_SUCCESS );
} /* chat_write_msg */

int
chat_write( CHAT_ctx c, int type, int fd, const void *buf,
            size_t buflen, size_t *r_nwritten )
{
    packet_s pkt;
    size_t nwritten = 0;
    int rc = 0;
	
    if ( !c || !buf || !buflen )
        return set_lasterr( CCERR_INV_VALUE );
			
    memset( &pkt, 0, sizeof pkt );
    pkt.type = type;
    pkt.data_size = buflen;
    memcpy( pkt.data, buf, buflen );
		
    rc = write_packet( c, &pkt, fd, &nwritten );
    if ( rc )
        return rc;
    if ( r_nwritten )
        *r_nwritten = nwritten;
  
    return set_lasterr( CCERR_SUCCESS );
} /* chat_write */

int
chat_read_msg( CHAT_ctx c, int fd, void *buf, size_t buflen, size_t *r_nread )
{
    packet_s pkt;
    u8 r_mdc[20];
    size_t nread;
    int rc = 0;
	
    if ( !c || !buf || !buflen )
        return set_lasterr( CCERR_INV_VALUE );

    if ( r_nread )
        *r_nread = 0;
    memset( &pkt, 0, sizeof pkt );
    rc = read_packet( c, &pkt, fd, &nread );
    if ( rc || pkt.type != PACKET_DATA )
        return rc;
    memcpy( buf, pkt.data, pkt.data_size );
    
    if ( pkt.use_mdc ) {
        gcry_md_hash_buffer( GCRY_MD_RMD160, r_mdc, buf, pkt.data_size );
        if ( memcmp( pkt.mdc, r_mdc, 20 ) )
            return CCERR_MDC;
    }
    if ( r_nread )
        *r_nread = nread;
	
    return set_lasterr( CCERR_SUCCESS );
} /* chat_read_msg */

int
chat_read( CHAT_ctx c, int *r_type, int fd, void *buf, size_t buflen,
           size_t *r_nread )
{
    packet_s pkt;
    size_t nread;
    int rc = 0;
	
    if ( !c || !buf || !buflen )
        return set_lasterr( CCERR_INV_VALUE );

    if ( r_nread )
        *r_nread = 0;
    memset( &pkt, 0, sizeof pkt );
    rc = read_packet( c, &pkt, fd, &nread );
    if ( rc )
        return rc;
    if ( r_type )
        *r_type = pkt.type;
    memcpy( buf, pkt.data, pkt.data_size );
    if ( r_nread )
        *r_nread = nread;
	
    return set_lasterr( CCERR_SUCCESS );
} /* chat_read */

static GCRY_MPI
sexp_to_mpi( GCRY_SEXP sexp, const char *val )
{
    GCRY_SEXP l = NULL;
    GCRY_MPI m = NULL;

    l = gcry_sexp_find_token( sexp, val, 0 );
    if ( !l ) {
        set_lasterr( CCERR_GCRYPT );
        return NULL;
    }
    m = gcry_sexp_nth_mpi( l, 1, 0 );
    if ( !m ) {
        set_lasterr( CCERR_GCRYPT );
        return NULL;
    }
    set_lasterr( CCERR_SUCCESS );
    return m;
} /* sexp_to_mpi */

#if 0 /* debug code */
void
hexdump( const u8 *buf, size_t buflen )
{
    int i = 0;
    wprintw( msg, "(%d) ", buflen );
    for ( i = 0; i < buflen ; i++ ) {
        if ( i == (buflen-1) )
            wprintw( msg, "%02x", buf[i] );
        else
            wprintw( msg,"%02x:", buf[i] );
    }
    wprintw( msg,"\n" );
    wrefresh( msg );
}
#endif

int
pk_sign( public_key_s *pk, const u8 *md, size_t mdlen, u8 *ret_r, u8 *ret_s )
{
    GCRY_SEXP s_md = NULL, s_sk = NULL, s_sig = NULL;
    GCRY_MPI m, mr = NULL, ms = NULL;
    int rc = 0, nbytes = 0, dlen = 0;

    if ( opt.disable_pk ) {
        /* public key authentication was disabled. we fill the buffers
           with values to prevent that scrambled data is used. */
        memset( ret_r, 'r', 20 );
        memset( ret_s, 's', 20 );
        return set_lasterr( CCERR_SUCCESS );
    }
    
    if ( !pk )
        return set_lasterr( CCERR_NOKEY );

    dlen = gcry_md_get_algo_dlen( GCRY_MD_SHA1 );
    if ( !pk->secret || mdlen != dlen )
        return set_lasterr( CCERR_GENERAL );
    rc = gcry_mpi_scan( &m, GCRYMPI_FMT_USG, md, &mdlen );
    if ( !rc ) {
        rc = gcry_sexp_build( &s_md, NULL, "%m", m );
        gcry_mpi_release( m );
    }
    if ( !rc ) {
        const char *fmt = "(private-key(dsa(p%m)(q%m)(g%m)(y%m)(x%m)))";
        rc = gcry_sexp_build( &s_sk, NULL, fmt, pk->p, pk->q, pk->g, pk->y,
                              pk->x );
    }
    if ( !rc )
        rc = gcry_pk_sign( &s_sig, s_md, s_sk );
    if ( !rc ) {
        mr = sexp_to_mpi( s_sig, "r" );
        ms = sexp_to_mpi( s_sig, "s" );
        nbytes = dlen; gcry_mpi_print( GCRYMPI_FMT_USG, ret_r, &nbytes, mr );
        nbytes = dlen; gcry_mpi_print( GCRYMPI_FMT_USG, ret_s, &nbytes, ms );
        gcry_mpi_release( mr );
        gcry_mpi_release( ms );
    }
    gcry_sexp_release( s_md );
    gcry_sexp_release( s_sk );
    gcry_sexp_release( s_sig );

    if ( rc )
        return set_lasterr( CCERR_GCRYPT );
    return set_lasterr( CCERR_SUCCESS );
} /* pk_sign */

int
pk_verify( public_key_s *pk, const u8 *r, const u8 *s,
           const u8 *md, size_t mdlen )
{
    GCRY_SEXP s_pk = NULL, s_sig = NULL, s_md = NULL;
    GCRY_MPI m = NULL, mr = NULL, ms = NULL;
    const char *fmt;
    int rc = 0, nbytes = 0, dlen = 0;

    if ( opt.disable_pk ) {
        /* we always return '0' to indicate a success */
        return set_lasterr( CCERR_SUCCESS );
    }
    
    if ( !pk )
        return set_lasterr( CCERR_NOKEY );

    dlen = gcry_md_get_algo_dlen( GCRY_MD_SHA1 );
    nbytes = dlen;
    rc = gcry_mpi_scan( &mr, GCRYMPI_FMT_USG, r, &nbytes );
    if ( !rc ) {
        nbytes = dlen;
        rc = gcry_mpi_scan( &ms, GCRYMPI_FMT_USG, s, &nbytes );
    }
    rc = gcry_mpi_scan( &m, GCRYMPI_FMT_USG, md, &mdlen );
    if ( !rc ) {
        rc = gcry_sexp_build( &s_md, NULL, "%m", m );
        gcry_mpi_release( m );   
    }
    if ( !rc ) {
        fmt = "(public-key(dsa(p%m)(q%m)(g%m)(y%m)))";
        rc = gcry_sexp_build( &s_pk, NULL, fmt, pk->p, pk->q, pk->g, pk->y );
    }
    if ( !rc ) {
        fmt = "(sig-val(dsa(r%m)(s%m)))";
        rc = gcry_sexp_build( &s_sig, NULL, fmt, mr, ms );
        gcry_mpi_release( mr );
        gcry_mpi_release( ms );
    }
    if ( !rc )
        rc = gcry_pk_verify( s_sig, s_md, s_pk );
    
    gcry_sexp_release( s_pk );
    gcry_sexp_release( s_md );
    gcry_sexp_release( s_sig );
    
    if ( rc )
        return set_lasterr( CCERR_GCRYPT );
    return set_lasterr( CCERR_SUCCESS );
} /* pk_verify */
