/* screen.c - (N)CURSES routines.  
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ncurses.h>
#ifdef __MINGW32__
 #include <windows.h>
#endif
#include <gcrypt.h>
#ifdef HAVE_TERMIOS_H
 #include <termios.h>
#endif

#include "context.h"
#include "gnuzza.h"

#ifdef __MINGW32__
 static HANDLE conin = INVALID_HANDLE_VALUE;
 #define DEF_INPMODE ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT \
         | ENABLE_PROCESSED_INPUT
#endif

WINDOW *msg, *msgb;
WINDOW *inp, *inpb;

int
create_main_wnds( void )
{
    msgb = newwin( LINES - 5, COLS - 1, 0, 0 ); 
    msg = newwin( LINES - 7, COLS - 4, 1, 1 );
	
    inpb = newwin( 4, COLS, LINES - 4, 0 );
    inp = newwin( 1, COLS - 2, LINES - 2, 1 );
    
    idlok( msgb, TRUE );
    scrollok( msgb, TRUE );
    idlok( msg, TRUE );
    scrollok( msg, TRUE );
	
    idlok( inp, TRUE );
    idcok( inp, TRUE );
    scrollok( inp, TRUE );
	
    wborder( msgb, 0, 0, 0, 0, 0, 0, 0, 0 );
    wrefresh( msgb );
    wborder( inpb, 0, 0, 0, 0, 0, 0, 0, 0 );
    wrefresh( inpb );
	
#ifndef __MINGW32__
    keypad( inp, TRUE );
#endif
	
    return 0;
} /* create_main_wnds */

void
init_colors( void )
{
    init_pair( WHITE, COLOR_WHITE, COLOR_BLACK );
    init_pair( RED, COLOR_RED, COLOR_BLACK );
    init_pair( GREEN, COLOR_GREEN, COLOR_BLACK );
    init_pair( BLUE, COLOR_BLUE, COLOR_BLACK );
    init_pair( YELLOW, COLOR_YELLOW, COLOR_BLACK );
    init_pair( CYAN, COLOR_CYAN, COLOR_BLACK );
} /* init_colors */

int
tty_init_con( void )
{
#ifdef __MINGW32__
    SECURITY_ATTRIBUTES sec_attr;

    memset( &sec_attr, 0, sizeof sec_attr );
    sec_attr.nLength = sizeof(sec_attr);
    sec_attr.bInheritHandle = TRUE;
    conin = CreateFile( "CONIN$", GENERIC_READ, FILE_SHARE_READ, &sec_attr,
                        OPEN_EXISTING, 0, 0 );
    if ( conin == INVALID_HANDLE_VALUE ) {
        printf( "could not open console for reading.\n" );
        return -1;   
    }
    SetConsoleMode( conin, DEF_INPMODE );
#endif
    return 0;
} /* tty_init_con */

void
tty_deinit_con( void )
{
#ifdef __MINGW32__
    if ( conin != INVALID_HANDLE_VALUE ) {
        CloseHandle( conin );
        conin = INVALID_HANDLE_VALUE;
    }
#endif
} /* tty_deinit_con */

int
tty_read_passphrase( char *pass, size_t plen, int mode )
{
    #ifdef HAVE_TERMIOS_H
    struct termios term, termsave;
    #endif
    #ifdef __MINGW32__
    DWORD nread = 0, i = 0;
    char cbuf[1];
    #endif
    const char *s_conv = "Please enter a passphrase "
        "(empty for no passphrase): ",
        *s_repeat = "Repeat your passphrase: ",
        *s_norm = "Please enter your passphrase: ";

    switch ( mode ) {
    case 0:  fprintf( stdout, "%s", s_norm ); break;
    case 1: fprintf( stdout, "%s", s_conv ); break;
    case 2: fprintf( stdout, "%s", s_repeat ); break;
    default: fprintf( stdout, "%s", s_norm ); break;
    }
    
    #ifdef HAVE_TERMIOS_H
    if ( tcgetattr( 0, &termsave ) ) {
        printf( "tcgetattr() failed: %s\n", strerror(errno) );
        return -1;
    }
    term = termsave;
    term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL );
    if ( tcsetattr( 0, TCSAFLUSH, &term ) ) {
        printf( "tcsetattr() failed: %s\n", strerror(errno) );
        return -1;
    }
    #endif
    #ifdef __MINGW32__
    SetConsoleMode( conin, ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT );
    for ( ;; ) {
        if ( !ReadConsole( conin, cbuf, 1, &nread, NULL ) ) {
            printf( "could not read console.\n" );
            SetConsoleMode( conin, DEF_INPMODE );
            return -1;
        }
        if ( !nread )
            continue;
        if ( *cbuf == '\n' ) {
            pass[i++] = cbuf[0]; /* we remove it later */
            break;
        }
        pass[i++] = cbuf[0];
    }
    SetConsoleMode( conin, DEF_INPMODE );
    #else
    fgets( pass, plen, stdin );
    #endif
    pass[strlen(pass)-1] = '\0';

    #ifdef HAVE_TERMIOS_H
    if ( tcsetattr( 0, TCSAFLUSH, &termsave ) ) {
        printf( "tcsetattr() failed: %s\n", strerror(errno) );
        return -1;
    }
    #endif
    fputc( '\n', stdout );
    return 0;
} /* tty_read_passphrase */

