/* sftp.c - Secure file transfer
 *   Copyright (C) 2002 Timo Schulz <ts@winpt.org>
 *
 * This file is part of Gnuzza.
 *
 * Gnuzza is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Gnuzza is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
#ifdef __MINGW32__
 #include <windows.h>
#endif
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <ncurses.h>
#include <gcrypt.h>

#include "types.h"
#include "context.h"
#include "sftp.h"
#include "socket.h"

static FILE *sfp = NULL;
static int sftp_debug = 0;

#define GET32( buf, idx ) ((buf[(idx)] << 24)|(buf[(idx)+1] << 16) \
                          |(buf[(idx)+2] << 8)|(buf[(idx)+3]))
#define PUT32( buf, val, idx ) buf[(idx)] = (val) >> 24; \
                               buf[(idx)+1] = (val) >> 16; \
                               buf[(idx)+2] = (val) >>  8; \
                               buf[(idx)+3] = (val) >>  0

const char*
sftp_strerror( int ec )
{
    switch ( ec ) {
        /* protocol id's */
    case SSH_FX_OK: return "operation successfull";
    case SSH_FX_EOF: return "end-of-file";
    case SSH_FX_NO_SUCH_FILE: return "no such file or directory";
    case SSH_FX_PERMISSION_DENIED: return "permission denied";
    case SSH_FX_FAILURE: return "operation failure";
    case SSH_FX_BAD_MESSAGE: return "badly formatted packet";
        /* program id's */
    case SFTP_ERR_MEM: return "memory error";
    case SFTP_ERR_PKT: return "invalid packet";
    case SFTP_ERR_OPEN: return "could not open file";
    case SFTP_ERR_READ: return "could not read channel";
    case SFTP_ERR_WRITE: return "could not write channel";
    case SFTP_ERR_CLOSE: return "could not close handle";
    case SFTP_ERR_PROTO: return "protocols does not match";
    default: return "unknown error";
    }
    return NULL;
} /* sftp_strerror */

static int
errno_to_sftp( int ec, FILE *fp )
{
    if ( fp && feof( fp ) )
        return SSH_FX_EOF;
    switch ( ec ) {
    case 0: return SSH_FX_OK;
    case ENOENT:
    case ENOTDIR: return SSH_FX_NO_SUCH_FILE;
    case EPERM:
    case EACCES: return SSH_FX_PERMISSION_DENIED;
    case EINVAL: return SSH_FX_BAD_MESSAGE;
    default: return SSH_FX_FAILURE;
    }
} /* errno_to_sftp */

static void
cipher_stream( conn_struct *c, int encrypt, u8 *buf, size_t buflen )
{
    GCRY_CIPHER_HD hd = NULL;

    if ( encrypt )
        hd = c->is_server ? c->srv : c->cli;
    else
        hd = !c->is_server ? c->srv : c->cli;
    if ( hd )
        gcry_cipher_encrypt( hd, buf, buflen, NULL, 0 );
} /* cipher_stream */

static int
write_packet( conn_struct *conn, int type, bstring_struct *data )
{
    u8 *pkt = NULL;
    int nwritten = 0, pktlen = 0;

    if ( sftp_debug > 1 )
        fprintf( sfp, "s: fd%d type %d len %d\n", conn->hfd, type, data->len );
    if ( data->len > MAX_PKT_SIZE ) {
        if ( sftp_debug )
            fprintf( sfp, "send: packet size too big %d.\n", data->len );
        return SFTP_ERR_PKT;   
    }
    pktlen = data->len + 4 + 1;
    pkt = calloc( 1, pktlen + 1 );
    if ( pkt == NULL )
        return SFTP_ERR_MEM;
    PUT32( pkt, data->len + 1, 0 );
    pkt[4] = type;
    memcpy( pkt + 5, data->d, data->len );
    cipher_stream( conn, 1, pkt, pktlen );
    nwritten = sock_write( conn->hfd, pkt, pktlen );
    free( pkt );
    if ( nwritten == -1 )
        return SFTP_ERR_WRITE;
    return 0;
} /* write_packet */

static int
read_packet( conn_struct *conn, bstring_struct **ret_data )
{
    u8 buf[5];
    bstring_struct *data = NULL;
    int nread = 0, dlen = 0, type = 0;

    nread = sock_read( conn->hfd, buf, 5 );
    if ( nread == -1 )
        return SFTP_ERR_READ;
    if ( nread == 0 )
        return 0; /* EOF */
    cipher_stream( conn, 0, buf, 5 );
    dlen = GET32( buf, 0 );
    type = buf[4];
    if ( sftp_debug > 1 )
        fprintf( sfp, "r: fd%d type %d len %d.\n", conn->hfd, type, dlen );
    if ( dlen > MAX_PKT_SIZE ) {
        if ( sftp_debug )
            fprintf( sfp, "recv: packet size too big %d.\n", dlen );
        return SFTP_ERR_PKT;
    }
    data = calloc( 1, sizeof *data + dlen - 1 );
    if ( data == NULL )
        return SFTP_ERR_MEM;
    nread = sock_read( conn->hfd, data->d, dlen - 1 );
    if ( nread == -1 )
        return SFTP_ERR_READ;
    data->len = dlen - 1;
    cipher_stream( conn, 0, data->d, dlen - 1 );
    if ( ret_data )
        *ret_data = data;
    return type;
} /* read_packet */

static int
do_send_init( conn_struct *conn, int version )
{
    bstring_struct ver, *pkt = NULL;
    int rc = 0, proto_ver = 0;

    ver.len = 4;
    PUT32( ver.d, version, 0 );
    rc = write_packet( conn, SSH_FXP_INIT, &ver );
    if ( rc )
        return rc;
    if ( read_packet( conn, &pkt ) != SSH_FXP_VERSION ) {
        rc = SFTP_ERR_PKT;
        goto fail;
    }
    proto_ver = GET32( pkt->d, 0 );
    if ( proto_ver != version )
        rc = SFTP_ERR_PROTO;
fail:
    free( pkt );
    return rc;
} /* do_send_init */

static int
do_proc_init( conn_struct *conn, int proto_ver, bstring_struct *pkt )
{
    bstring_struct ver;
    int version = 0, rc = 0;

    if ( pkt->len != 4 )
        return SFTP_ERR_PKT;
    version = GET32( pkt->d, 0 );
    ver.len = 4;
    PUT32( ver.d, proto_ver, 0 );
    rc = write_packet( conn, SSH_FXP_VERSION, &ver );
    if ( rc )
        return rc;
    if ( version != proto_ver )
        return SFTP_ERR_PROTO;
    return 0;
} /* do_proc_init */

static int
do_send_status( conn_struct *conn, int ec )
{
    const char *s = sftp_strerror( ec );
    bstring_struct pkt;
    u32 slen = 0, pos = 0;

    if ( sftp_debug )
        fprintf( sfp, "status: %s\n", s );
    slen = strlen( s );
    PUT32( pkt.d, conn->id, 0 ); pos += 4;
    PUT32( pkt.d, ec, pos ); pos += 4;
    PUT32( pkt.d, slen, pos ); pos += 4;
    memcpy( pkt.d + pos, s, slen ); pos += slen;
    pkt.len = pos;
    return write_packet( conn, SSH_FXP_STATUS, &pkt );
} /* do_send_status */

static int
do_proc_status( conn_struct *conn, bstring_struct *pkt )
{
    int ec = 0;

    if ( pkt->len > MAX_PKT_SIZE )
        return SFTP_ERR_PKT;
    ec = GET32( pkt->d, 4 );
    if ( sftp_debug > 1 )
        fprintf( sfp, "status: %s.\n", sftp_strerror( ec ) );
    
    return ec;
} /* do_proc_status */

static int
do_send_handle( conn_struct *conn, handle_struct *hd )
{
    bstring_struct pkt;
    u32 slen = 0, pos = 0;
    
    slen = strlen( hd->handle );
    PUT32( pkt.d, conn->id, 0 ); pos += 4;
    PUT32( pkt.d, slen, pos ); pos += 4;
    memcpy( pkt.d + pos, hd->handle, slen ); pos += slen;
    pkt.len = pos;
    return write_packet( conn, SSH_FXP_HANDLE, &pkt );
} /* do_send_handle */

static int
do_proc_handle( conn_struct *conn, handle_struct *hd, bstring_struct *pkt )
{
    u32 id = 0, slen = 0, pos = 0;
    
    if ( pkt->len > MAX_PKT_SIZE )
        return SFTP_ERR_PKT;
    id = GET32( pkt->d, 0 ); pos += 4;
    slen = GET32( pkt->d, pos ); pos += 4;
    hd->handle = calloc( 1, slen + 1 );
    if ( hd->handle == NULL )
        return SFTP_ERR_MEM;
    strncpy( hd->handle, pkt->d + pos, slen );
    return 0;
} /* do_proc_handle */

static int
do_send_open( conn_struct *conn, const char *name, size_t flags, size_t attr,
              handle_struct **ret_hd )
{
    bstring_struct op, *pkt = NULL;
    handle_struct *hd = NULL;
    u32 id = 0, slen = 0, pos = 0;
    int rc = 0;

    id = conn->id++;
    slen = strlen( name );
    PUT32( op.d, id, 0 ); pos += 4;
    PUT32( op.d, slen, pos ); pos += 4;
    memcpy( op.d + pos, name, slen ); pos += slen;
    PUT32( op.d, flags, pos ); pos += 4;
    op.len = pos;
    rc = write_packet( conn, SSH_FXP_OPEN, &op );
    if ( rc )
        return rc;
    rc = read_packet( conn, &pkt );
    if ( rc == SSH_FXP_STATUS ) {
        free( pkt );
        return do_proc_status( conn, pkt );
    }
    else if ( rc == SSH_FXP_HANDLE ) {
        hd = calloc( 1, sizeof *hd );
        if ( hd == NULL ) {
            rc = SFTP_ERR_MEM;
            goto fail;
        }
        rc = do_proc_handle( conn, hd, pkt );
        if ( rc ) {
            free( hd );
            goto fail;
        }
        if ( ret_hd )
            *ret_hd = hd;
    }
    
fail:
    free( pkt );
    return rc;
} /* do_send_open */

static int
file_exist( const char *file )
{
    FILE *fp = fopen( file, "r" );
    if ( fp ) {
        fclose( fp );
        return -1;
    }
    return 0;
} /* file_exist */
    
static int
do_proc_open( conn_struct *conn, handle_struct **ret_hd, bstring_struct *pkt )
{
    handle_struct *hd;
    u32 id = 0, slen = 0;
    int flags = 0, pos = 0;

    if ( pkt->len > MAX_PKT_SIZE )
        return SFTP_ERR_PKT;
    conn->id++;
    hd = calloc( 1, sizeof *hd );
    if ( hd == NULL )
        return SFTP_ERR_MEM;
    id = GET32( pkt->d, 0 ); pos += 4;
    slen = GET32( pkt->d, pos ); pos += 4;
    hd->handle = calloc( 1, slen + 9 + 1 );
    if ( hd->handle == NULL ) {
        free( hd );
        return SFTP_ERR_MEM;
    }
    strncpy( hd->handle, pkt->d + pos, slen ); pos += slen;
    flags = GET32( pkt->d, pos ); pos += 4;
    if ( flags != SSH_FXF_CREAT ) {
        free( hd );
        return SFTP_ERR_PKT;
    }
    /* if the file already exists, we use the format file.%id. due to the
     * fact that the id changed with every open request we don't need to care.
     */
    if ( file_exist( hd->handle ) ) {
        char n[32]; sprintf( n, ".%d", (int)conn->id );
        strcat( hd->handle, n );
    }
    hd->fp = fopen( hd->handle, "w+b" );
    if ( hd->fp == NULL ) {
        free( hd );
        do_send_status( conn, errno_to_sftp( errno, NULL ) );
        return SFTP_ERR_OPEN;
    }
    if ( ret_hd )
        *ret_hd = hd;
    return do_send_handle( conn, hd );
} /* do_proc_open */

static int
do_send_close( conn_struct *conn, handle_struct *hd )
{
    bstring_struct cl, *pkt = NULL;
    u32 id = 0, slen = 0, pos = 0;
    int rc = 0;

    id = conn->id++;
    slen = strlen( hd->handle );
    PUT32( cl.d, id, 0 ); pos += 4;
    PUT32( cl.d, slen, 4 ); pos += 4;
    memcpy( cl.d + pos, hd->handle, slen ); pos += slen;
    cl.len = pos;
    rc =  write_packet( conn, SSH_FXP_CLOSE, &cl );
    if ( rc )
        return rc;
    rc = read_packet( conn, &pkt );
    if ( rc != SSH_FXP_STATUS ) {
        rc = SFTP_ERR_PKT;
        goto fail;
    }
    
    rc = do_proc_status( conn, pkt );
    switch ( rc ) {
    case SSH_FX_OK: rc = 0; break;
    default: rc = SFTP_ERR_CLOSE; break;
    }
    
fail:
    free( pkt );
    return rc;
} /* do_send_close */

static int
do_proc_close( conn_struct *conn, handle_struct *hd, bstring_struct *pkt )
{
    char *handle = NULL;
    u32 id = 0, slen = 0, pos = 0;
    int rc = 0;

    if ( pkt->len > MAX_PKT_SIZE )
        return SFTP_ERR_PKT;
    conn->id++;
    id = GET32( pkt->d, 0 ); pos += 4;
    slen = GET32( pkt->d, 4 ); pos += 4;
    handle = calloc( 1, slen + 1 );
    if ( handle == NULL )
        return SFTP_ERR_MEM;
    strncpy( handle, pkt->d + pos, slen );
    if ( hd->is_used ) {
        int err = 0;
        if ( sftp_debug )
            fprintf( sfp, "%s: close hd.\n", hd->handle );
        if ( fclose( hd->fp ) == -1 )
            err = errno;
        rc = errno_to_sftp( err, NULL );   
        hd->is_used = 0;
    }
    free( handle );
    return do_send_status( conn, rc );
} /* do_proc_close */

static int
do_send_write( conn_struct *conn, handle_struct *hd, u64 off,
               const u8 *buf, size_t buflen )
{
    bstring_struct wr, *pkt = NULL;
    u32 id = 0, pos = 0;
    int rc = 0;

    id = conn->id++;
    PUT32( wr.d, id, 0 ); pos += 4;
    PUT32( wr.d, (off >> 32), pos ); pos += 4;
    PUT32( wr.d, off, pos ); pos += 4;
    PUT32( wr.d, buflen, pos ); pos += 4;
    memcpy( wr.d + pos, buf, buflen ); pos += buflen;
    wr.len = pos;
    rc = write_packet( conn, SSH_FXP_WRITE, &wr );
    if ( rc )
        return rc;
    rc = read_packet( conn, &pkt );
    if ( rc != SSH_FXP_STATUS ) {
        free( pkt );
        return SFTP_ERR_PKT;
    }
    if ( do_proc_status( conn, pkt ) != SSH_FX_OK ) {
        free( pkt );
        return SFTP_ERR_WRITE;
    }
    free( pkt );
    return 0;
} /* do_send_write */

static int
do_proc_write( conn_struct *conn, handle_struct *hd, bstring_struct *pkt )
{
    u32 id = 0, off = 0, pos = 0;
    int dlen = 0, err = 0;
    u8 *buf = NULL;

    conn->id++;
    id = GET32( pkt->d, 0 ); pos += 4;
    pos += 4; /* skip hight 32 bit */
    off = GET32( pkt->d, pos ); pos += 4;
    dlen = GET32( pkt->d, pos ); pos += 4;
    buf = calloc( 1, dlen + 1 );
    if ( buf == NULL )
        return SFTP_ERR_MEM;
    memcpy( buf, pkt->d + pos, dlen );
    if ( fseek( hd->fp, off, SEEK_SET ) == -1 ) {
        err = EINVAL;
        goto fail;
    }
    if ( !fwrite( buf, 1, dlen, hd->fp ) )
        err = EINVAL;
    
fail:
    free( buf );
    return do_send_status( conn, errno_to_sftp( err, hd->fp ) );
} /* do_proc_write */

int
sftp_proc_packets( struct sftp_s *ftp, void (*cb)( const char *text ) )
{
    bstring_struct *pkt = NULL;
    handle_struct *hd = NULL;
    int rc = 0, type = 0;

    ftp->conn.id = 0;
    
    do {
        type = read_packet( &ftp->conn, &pkt );
        switch ( type ) {
        case SSH_FXP_INIT:
            rc = do_proc_init( &ftp->conn, SFTP_PROTO_VERSION, pkt );
            break;
            
        case SSH_FXP_OPEN:
            rc = do_proc_open( &ftp->conn, &hd, pkt );
            if ( !rc )
                hd->is_used = 1;
            break;
            
        case SSH_FXP_CLOSE:
            rc = do_proc_close( &ftp->conn, hd, pkt );
            if ( !rc && cb )
                cb( hd->handle );
            free( hd->handle ); hd->handle = NULL;
            free( hd ); hd = NULL;
            break;
            
        case SSH_FXP_WRITE:
            rc = do_proc_write( &ftp->conn, hd, pkt );
            break;
            
        default:
            if ( type < 0 )
                rc = type;
            goto fail;
        }
        free( pkt ); pkt = NULL;
    } while ( type > 0 );
fail:

    return rc;
} /* proc_packets */

int
sftp_send_file( struct sftp_s *ftp, const char *local, const char *remote )
{
    handle_struct *hd = NULL;
    FILE *fp;
    u8 buf[2048];
    u64 off = 0;
    int rc = 0;
    size_t nr = 0;

    fp = fopen( local, "rb" );
    if ( fp == NULL )
        return SFTP_ERR_OPEN;

    ftp->conn.id = 0;
    rc = do_send_init( &ftp->conn, SFTP_PROTO_VERSION );
    if ( rc )
        return rc;
    rc = do_send_open( &ftp->conn, remote, SSH_FXF_CREAT, 0, &hd );
    if ( rc )
        return rc;
    
    while ( !feof( fp ) ) {
        nr = fread( buf, 1, sizeof buf-1, fp );
        if ( !nr )
            break;
        if ( sftp_debug )
            fprintf( sfp, "fread fd%d bytes %d\n", fileno(fp), nr );
        rc = do_send_write( &ftp->conn, hd, off, buf, nr );
        if ( rc )
            break;
        off += nr;
    }
    rc = do_send_close( &ftp->conn, hd );
    free( hd ); hd = NULL;
    fclose( fp );
    
    return rc;
} /* sftp_send_file */
    
void
sftp_set_debug_level( int level )
{
    sftp_debug = level;
    if ( !sfp )
        sfp = stderr;
} /* sftp_set_debug_level */

void
sftp_set_debug_file( const char *file )
{
    sfp = fopen( file, "w+b" );
    if ( sfp == NULL )
        sfp = stderr;
} /* sftp_set_debug_file */
