/* socket.c - Socket operations.  
 *  Copyright (C) 2001, 2002 Timo Schulz <ts@winpt.org>
 *  
 * This file is part of Gnuzza.
 *  
 * Gnuzza is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * Gnuzza is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with Gnuzza; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */ 

#ifdef HAVE_CONFIG_H
 #include <config.h>
#endif
 #ifndef __MINGW32__
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
#else
 #include <windows.h>
#endif
#include <string.h>
#include <stdio.h>
#include <gcrypt.h>
#include <ncurses.h>

#include "types.h"
#include "errors.h"
#include "context.h"
#include "socket.h"
#include "gnuzza.h"
#include "crypto.h"

int
connect_to_server( const char *server, int port )
{
    int rc = 0, sock_fd = -1;
    struct sockaddr_in addr;
    struct hostent *hp;
    u32 iaddr = 0;

    if ( !server ) {
        set_lasterr( CCERR_INV_VALUE );
        return -1;
    }

    /* we are only verbose for the gnuzza connection. when we try
       to use the sftp subsystem, be quiet. */
    if ( port == PORT ) {
        wprintw( msg, "CONNECT_TO %s\n", server );
        wrefresh( msg );
    }
	
    memset( &addr, 0, sizeof addr );
    addr.sin_family = AF_INET;
    addr.sin_port = htons( port );
	
    if ( (iaddr = inet_addr( server ) ) != -1 )
        memcpy( &addr.sin_addr, &iaddr, sizeof iaddr );
    else if ( ( hp = gethostbyname( server ) ) ) {
        if ( hp->h_addrtype != AF_INET ) {
            wprintw( msg, "unsupported address family: must be TCP\n" );
            wrefresh( msg );
            set_lasterr( CCERR_CONNECT );
            return -1;
        }
        if ( hp->h_length != 4 ) {
            wprintw( msg, "unsupported address length: must be 4 bytes!\n" );
            wrefresh( msg );
            set_lasterr( CCERR_CONNECT );
            return -1;
        }
        memcpy( &addr.sin_addr, hp->h_addr, hp->h_length );
    }
		
    sock_fd = socket( PF_INET, SOCK_STREAM, 0 );
    if ( sock_fd == -1 ) {
        sock_error( "socket" );
        set_lasterr( CCERR_CONNECT );
        return -1;
    }
    
    rc = connect( sock_fd, (struct sockaddr *)&addr, sizeof addr );
    if ( rc == -1 ) {
        sock_error( "connect" );
        sock_close( sock_fd );
        set_lasterr( CCERR_CONNECT );
        return -1;
    }
	
    return sock_fd;
} /* connect_to_server */

int
wait_for_client( const char *device, int port, char **ret_cip )
{
    struct sockaddr_in srv_addr, cli_addr;
    int listen_fd = -1, conn_fd = -1;
    int rc = 0, cli_len, i = 0;
	
    listen_fd = socket( PF_INET, SOCK_STREAM, 0 );
    if ( listen_fd == -1 ) {
        sock_error( "socket" );
        return -1;
    }
	
    memset( &srv_addr, 0, sizeof srv_addr );
    srv_addr.sin_family = AF_INET;
    srv_addr.sin_addr.s_addr = device ? inet_addr( device ): INADDR_ANY;
    srv_addr.sin_port = htons( port );
	
    i = 1;
    setsockopt( listen_fd, SOL_SOCKET, SO_REUSEADDR, (u8*)&i, sizeof i );
	
    rc = bind( listen_fd, (struct sockaddr *)&srv_addr, sizeof srv_addr );
    if ( rc == -1 ) {
        sock_error( "bind" );
        return -1;
    }
	
    rc = listen( listen_fd, 5 );
    if ( rc == -1 ) {
        sock_error( "listen" );
        return -1;
    }
	
    cli_len = sizeof cli_addr;
    conn_fd = accept( listen_fd, (struct sockaddr *)&cli_addr, &cli_len );
    if ( conn_fd == -1 ) {
        sock_error( "accept" );
        return -1;
    }

    /* see notice in connect_to_server */ 
    if ( port == PORT ) {
        wprintw( msg, "CONNECT_FROM %s:%d\n",
                 inet_ntoa( cli_addr.sin_addr ),
                 htons( cli_addr.sin_port ) );
        wrefresh( msg );
    }
    
    if ( ret_cip )
        *ret_cip = gcry_strdup( inet_ntoa( cli_addr.sin_addr ) );
    
    return conn_fd;
} /* wait_for_client */

